import numpy as np
from joblib import (
    Parallel,
    delayed,
)

from ..linear_model import (
    OrthogonalRegression,
    RidgeRegression2FoldCV,
)
from ..model_selection import train_test_split
from ..preprocessing import StandardFlexibleScaler


def pointwise_global_reconstruction_error(
    X,
    Y,
    train_idx=None,
    test_idx=None,
    scaler=None,
    estimator=None,
):
    r"""Computes the pointwise global reconstruction error using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by linear regression:

    .. math::

        GRE^{(i)}(X,Y) = \min_W ||y_i - x_iW||

    If used with X and Y of shape (n_samples, n_features) it computes the pointwise
    global reconstruction error of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The error is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction error of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The error is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    pointwise_global_reconstruction_error : ndarray
        The global reconstruction error for each sample/point

    """
    (
        train_idx,
        test_idx,
        scaler,
        estimator,
    ) = check_global_reconstruction_measures_input(
        X, Y, train_idx, test_idx, scaler, estimator
    )
    X_train, X_test, Y_train, Y_test = (
        X[train_idx],
        X[test_idx],
        Y[train_idx],
        Y[test_idx],
    )

    scaler.fit(X_train)
    X_train = scaler.transform(X_train)
    X_test = scaler.transform(X_test)
    scaler.fit(Y_train)
    Y_train = scaler.transform(Y_train)
    Y_test = scaler.transform(Y_test)

    estimator.fit(X_train, Y_train)

    return np.linalg.norm(Y_test - estimator.predict(X_test), axis=1)


def global_reconstruction_error(
    X,
    Y,
    test_idx=None,
    train_idx=None,
    scaler=None,
    estimator=None,
):
    r"""Computes the global reconstruction error using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by linear regression:

    .. math::

        GRE(X,Y) = \min_W ||Y - XW||

    If used with X and Y of shape (n_samples, n_features) it computes the
    global reconstruction error of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The error is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction error of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The error is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    global_reconstruction_error : ndarray
        The global reconstruction error

    """
    pointwise_global_reconstruction_error_values = (
        pointwise_global_reconstruction_error(
            X,
            Y,
            train_idx=train_idx,
            test_idx=test_idx,
            scaler=scaler,
            estimator=estimator,
        )
    )
    return np.linalg.norm(pointwise_global_reconstruction_error_values) / np.sqrt(
        len(pointwise_global_reconstruction_error_values)
    )


def pointwise_global_reconstruction_distortion(
    X,
    Y,
    test_idx=None,
    train_idx=None,
    scaler=None,
    estimator=None,
):
    r"""Computes the pointwise global reconstruction distortion using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by orthogonal regression:

    .. math::

        GRD^{(i)}(X,Y) = \min_Q ||y_i - x_iQ\|| \quad\mathrm{subject\ to}\quad Q^TQ=I

    If used with X and Y of shape (n_samples, n_features) it computes the pointwise
    global reconstruction distortion of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The distortion is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction distortion of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The distortion is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    pointwise_global_reconstruction_distortion : ndarray
        The global reconstruction distortion for each sample/point

    """
    (
        train_idx,
        test_idx,
        scaler,
        estimator,
    ) = check_global_reconstruction_measures_input(
        X, Y, train_idx, test_idx, scaler, estimator
    )
    X_train, X_test, Y_train, Y_test = (
        X[train_idx],
        X[test_idx],
        Y[train_idx],
        Y[test_idx],
    )

    scaler.fit(X_train)
    X_train = scaler.transform(X_train)
    X_test = scaler.transform(X_test)
    scaler.fit(Y_train)
    Y_train = scaler.transform(Y_train)
    Y_test = scaler.transform(Y_test)

    predictions_Y_test = estimator.fit(X_train, Y_train).predict(X_test)
    orthogonal_predictions_Y_test = (
        OrthogonalRegression(use_orthogonal_projector=False)
        .fit(X_train, estimator.predict(X_train))
        .predict(X_test)
    )

    return np.linalg.norm(predictions_Y_test - orthogonal_predictions_Y_test, axis=1)


def global_reconstruction_distortion(
    X,
    Y,
    test_idx=None,
    train_idx=None,
    scaler=None,
    estimator=None,
):
    r"""Computes the global reconstruction distortion using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by orthogonal regression:

    .. math::

        GRD(X,Y) = \min_Q ||y - XQ\|| \quad\mathrm{subject\ to}\quad Q^TQ=I

    If used with X and Y of shape (n_samples, n_features) it computes the
    global reconstruction distortion of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The distortion is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction distortion of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The distortion is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    global_reconstruction_distortion : ndarray
        The global reconstruction distortion

    """
    pointwise_global_reconstruction_distortion_values = (
        pointwise_global_reconstruction_distortion(
            X,
            Y,
            train_idx=train_idx,
            test_idx=test_idx,
            scaler=scaler,
            estimator=estimator,
        )
    )
    return np.linalg.norm(pointwise_global_reconstruction_distortion_values) / np.sqrt(
        len(pointwise_global_reconstruction_distortion_values)
    )


def pointwise_local_reconstruction_error(
    X,
    Y,
    n_local_points,
    test_idx=None,
    train_idx=None,
    scaler=None,
    estimator=None,
    n_jobs=None,
):
    r"""Computes the pointwise local reconstruction error using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by linear regression:

    .. math::

        \tilde{\mathbf{x}}'_i = \bar{\mathbf{x}} + (\mathbf{x}_i
                                - \bar{\mathbf{x}})\mathbf{P}^{(i)}

    .. math::

        LRE^{(i)}(X,Y) = \|\mathbf{x}'_i - \tilde{\mathbf{x}}'_i\|^2

    If used with X and Y of shape (n_samples, n_features) it computes the pointwise
    local reconstruction error of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The error is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction error of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The error is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    n_local_points : int,
        Number of neighbour points used to compute the local reconstruction weight for
        each sample/point.

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    pointwise_local_reconstruction_error : ndarray
        The local reconstruction error for each sample/point

    """
    (
        train_idx,
        test_idx,
        scaler,
        estimator,
    ) = check_local_reconstruction_measures_input(
        X, Y, n_local_points, train_idx, test_idx, scaler, estimator
    )
    X_train, X_test, Y_train, Y_test = (
        X[train_idx],
        X[test_idx],
        Y[train_idx],
        Y[test_idx],
    )

    scaler.fit(X_train)
    X_train = scaler.transform(X_train)
    X_test = scaler.transform(X_test)
    scaler.fit(Y_train)
    Y_train = scaler.transform(Y_train)
    Y_test = scaler.transform(Y_test)

    squared_dist = (
        np.sum(X_train**2, axis=1)
        + np.sum(X_test**2, axis=1)[:, np.newaxis]
        - 2 * X_test @ X_train.T
    )

    n_test = X_test.shape[0]

    def local_reconstruction_error_i(i):
        # comments correspond notation in [gfrm]_
        local_env_idx = np.argsort(squared_dist[i])[:n_local_points]
        # D_{k-neigh}^{(i)}
        local_X_train = X_train[local_env_idx]
        # \bar{x}_F
        local_X_train_mean = np.mean(X_train[local_env_idx], axis=0)
        # D_{k-neigh}^{(i)}
        local_Y_train = Y_train[local_env_idx]
        # \bar{x}_F'
        local_Y_train_mean = np.mean(Y_train[local_env_idx], axis=0)
        # P_{FF'}
        estimator.fit(
            local_X_train - local_X_train_mean,
            local_Y_train - local_Y_train_mean,
        )
        # \tilde{x}_i' = \bar{x}_{F'} + (x_i - \bar{x}_F)P_{FF'}
        tilde_x_i_dash_test = local_Y_train_mean + estimator.predict(
            X_test[i, :][np.newaxis, :] - local_X_train_mean
        )
        # \|x_i' -  \tilde{x}_i'\|
        return np.linalg.norm(Y_test[i, :][np.newaxis, :] - tilde_x_i_dash_test)

    pointwise_local_reconstruction_error_values = np.array(
        Parallel(n_jobs=n_jobs)(
            delayed(local_reconstruction_error_i)(i) for i in range(n_test)
        )
    )
    return pointwise_local_reconstruction_error_values


def local_reconstruction_error(
    X,
    Y,
    n_local_points,
    test_idx=None,
    train_idx=None,
    scaler=None,
    estimator=None,
    n_jobs=None,
):
    r"""Computes the local reconstruction error using the source X
    to reconstruct the features or samples of target Y based on a minimization
    by linear regression:

    .. math::

        LRE(X,Y) = \sqrt{\sum_i LRE^{(i)}(X,Y)}/\sqrt{n_\text{test}}

    If used with X and Y of shape (n_samples, n_features) it computes the
    local reconstruction error of the features as defined in Ref. [Goscinski2021]_.
    In this case the number of samples of X and Y should agree with each other,
    but the number of features can be different. The error is expressed per sample.

    If used with X and Y of shape(n_features, n_samples) it computes the
    reconstruction error of the samples. In this case the number of
    features of X and Y should agree with each other, but the number of
    samples can be different. The error is expressed per feature.

    The default parameters mimics the ones of Ref. [Goscinski2021]_.

    Parameters
    ----------
    X : ndarray of shape (n_samples, X_n_features)
        Source data which reconstructs target Y.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    Y : ndarray of shape (n_samples, Y_n_targets)
        Target data which is reconstructed with X.
        For feature reconstruction of Y using X use input shape (samples, features).
        For sample reconstruction of Y using X use input shape (features, samples).

    n_local_points : int,
        Number of neighbour points used to compute the local reconstruction weight for
        each sample/point.

    train_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``test_idx`` is used. If ``train_size`` is
        also None, 2-fold split is taken.

    test_idx : ndarray, dtype=int, default=None
        array of indices used for training, if None,
        If None, the complement of the ``train_idx`` is used. If ``test_size`` is
        also None, 2-fold split is taken.

    scaler : object implementing fit/transfom
        Scales the X and Y before computing the reconstruction measure.
        The default value scales the features such that the reconstruction
        measure on the training set is upper bounded to 1.

    estimator : object implementing fit/predict, default=None
        Sklearn estimator used to reconstruct features/samples.

    Returns
    -------
    local_reconstruction_error : ndarray
        The local reconstruction error

    """
    pointwise_local_reconstruction_error_values = pointwise_local_reconstruction_error(
        X,
        Y,
        n_local_points,
        train_idx=train_idx,
        test_idx=test_idx,
        scaler=scaler,
        estimator=estimator,
        n_jobs=n_jobs,
    )
    return np.linalg.norm(pointwise_local_reconstruction_error_values) / np.sqrt(
        len(pointwise_local_reconstruction_error_values)
    )


def check_global_reconstruction_measures_input(
    X, Y, train_idx, test_idx, scaler, estimator
):
    """Returns default reconstruction measure inputs for all None parameters"""
    assert len(X) == len(Y)
    if (train_idx is None) and (test_idx is None):
        train_idx, test_idx = train_test_split(
            np.arange(len(X)),
            test_size=0.5,
            train_size=0.5,
            random_state=0x5F3759DF,
            shuffle=True,
            train_test_overlap=False,
        )
    elif train_idx is None:
        train_idx = np.setdiff1d(np.arange(len(X)), test_idx)
    elif test_idx is None:
        test_idx = np.setdiff1d(np.arange(len(X)), train_idx)

    if scaler is None:
        scaler = StandardFlexibleScaler()

    if estimator is None:
        estimator = RidgeRegression2FoldCV(
            alphas=np.geomspace(1e-9, 0.9, 20),
            alpha_type="relative",
            regularization_method="cutoff",
            random_state=0x5F3759DF,
            shuffle=True,
            scoring="neg_root_mean_squared_error",
            n_jobs=1,
        )
    return train_idx, test_idx, scaler, estimator


def check_local_reconstruction_measures_input(
    X, Y, n_local_points, train_idx, test_idx, scaler, estimator
):
    """Returns default reconstruction measure inputs for all None parameters"""
    # only needs to check one extra parameter
    assert len(X) >= n_local_points
    return check_global_reconstruction_measures_input(
        X, Y, train_idx, test_idx, scaler, estimator
    )



def get_progress_bar():
    """
    This function returns the appropriate version of tqdm, as determined by
    tqdm.auto. If tqdm is not installed, an ImportError is raised.
    """
    try:
        from tqdm.auto import tqdm

        return tqdm
    except ImportError:
        raise ImportError(
            "tqdm must be installed to use a progress bar."
            "Either install tqdm or re-run with"
            "progress_bar = False"
        )


def no_progress_bar(x):
    """
    This is the identity function, same as lambda x:x. It returns x.
    """
    return x



from ._reconstruction_measures import (
    check_global_reconstruction_measures_input,
    check_local_reconstruction_measures_input,
    global_reconstruction_distortion,
    global_reconstruction_error,
    local_reconstruction_error,
    pointwise_global_reconstruction_distortion,
    pointwise_global_reconstruction_error,
    pointwise_local_reconstruction_error,
)

__all__ = [
    "pointwise_global_reconstruction_error",
    "global_reconstruction_error",
    "pointwise_global_reconstruction_distortion",
    "global_reconstruction_distortion",
    "pointwise_local_reconstruction_error",
    "local_reconstruction_error",
    "check_global_reconstruction_measures_input",
    "check_local_reconstruction_measures_input",
]



import unittest

from skmatter.utils import get_progress_bar


class PBarTest(unittest.TestCase):
    def test_no_tqdm(self):
        """
        This test checks that the model cannot use a progress bar when tqdm
        is not installed
        """
        import sys

        sys.modules["tqdm"] = None

        with self.assertRaises(ImportError) as cm:
            _ = get_progress_bar()
        self.assertEqual(
            str(cm.exception),
            "tqdm must be installed to use a progress bar."
            "Either install tqdm or re-run with"
            "progress_bar = False",
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest
import numpy as np

from skmatter.datasets import (
    load_degenerate_CH4_manifold,
    load_csd_1000r,
    load_nice_dataset,
    load_who_dataset,
)


class NICEDatasetTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nice_data = load_nice_dataset()

    def test_load_nice_data(self):
        # test if representations and properties have commensurate shape
        self.assertTrue(
            self.nice_data.data.X.shape[0] == self.nice_data.data.y.shape[0]
        )
        self.assertTrue(self.nice_data.data.X.shape[0] == 500)
        self.assertTrue(self.nice_data.data.X.shape[1] == 160)
        self.assertTrue(len(self.nice_data.data.X.shape) == 2)

    def test_load_nice_data_descr(self):
        self.nice_data.DESCR


class DegenerateCH4Tests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.degenerate_CH4_manifold = load_degenerate_CH4_manifold()

    def test_load_degenerate_CH4_manifold_power_spectrum_shape(self):
        # test if representations have correct shape
        self.assertTrue(
            self.degenerate_CH4_manifold.data.SOAP_power_spectrum.shape == (162, 12)
        )

    def test_load_degenerate_CH4_manifold_bispectrum_shape(self):
        self.assertTrue(
            self.degenerate_CH4_manifold.data.SOAP_bispectrum.shape == (162, 12)
        )

    def test_load_degenerate_CH4_manifold_access_descr(self):
        self.degenerate_CH4_manifold.DESCR


class CSDTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.csd = load_csd_1000r()

    def test_load_csd_1000r_shape(self):
        # test if representations and properties have commensurate shape
        self.assertTrue(self.csd.data.X.shape[0] == self.csd.data.y.shape[0])

    def test_load_csd_1000r_access_descr(self):
        self.csd.DESCR


class WHOTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.size = 24240
        cls.shape = (2020, 12)
        cls.value = 5.00977993011475
        try:
            import pandas as pd  # noqa F401

            cls.has_pandas = True
            cls.who = load_who_dataset()
        except ImportError:
            cls.has_pandas = False

    def test_load_dataset_without_pandas(self):
        """
        Check if the correct exception occurs when pandas isn't present.
        """
        if self.has_pandas is False:
            with self.assertRaises(ImportError) as cm:
                _ = load_who_dataset()
            self.assertEqual(str(cm.exception), "load_who_dataset requires pandas.")

    def test_dataset_size_and_shape(self):
        """
        Check if the correct number of datapoints are present in the dataset.
        Also check if the size of the dataset is correct.
        """
        if self.has_pandas is True:
            self.assertEqual(self.who["data"].size, self.size)
            self.assertEqual(self.who["data"].shape, self.shape)

    def test_datapoint_value(self):
        """
        Check if the value of a datapoint at a certain location is correct.
        """
        if self.has_pandas is True:
            self.assertTrue(
                np.allclose(
                    self.who["data"]["SE.XPD.TOTL.GD.ZS"][1924], self.value, rtol=1e-6
                )
            )


if __name__ == "__main__":
    unittest.main()



"""
Sequential selection
"""

import numbers
import warnings
from abc import abstractmethod

import numpy as np
import scipy
from scipy.linalg import eigh
from scipy.sparse.linalg import eigsh
from sklearn.base import (
    BaseEstimator,
    MetaEstimatorMixin,
)
from sklearn.feature_selection._base import SelectorMixin
from sklearn.utils import (
    check_array,
    check_random_state,
    safe_mask,
)
from sklearn.utils._tags import _safe_tags
from sklearn.utils.validation import check_is_fitted

from .utils import (
    X_orthogonalizer,
    Y_feature_orthogonalizer,
    Y_sample_orthogonalizer,
    get_progress_bar,
    no_progress_bar,
    pcovr_covariance,
    pcovr_kernel,
)


class GreedySelector(SelectorMixin, MetaEstimatorMixin, BaseEstimator):
    """

    Transformer that adds, via greedy forward selection,
    features or samples to form a subset. At each stage, the model scores each
    feature or sample (without an estimator) and chooses that with the maximum score.

    Parameters
    ----------

    selection_type : str, {'feature', 'sample'}
        whether to choose a subset of columns ('feature') or rows ('sample').
        Stored in :py:attr:`self._axis_name` (as text) and :py:attr:`self._axis`
        (as 0 or 1 for 'sample' or 'feature', respectively).

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the features or samples are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining features. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------
    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected samples or features, for use in fitting
    y_selected_ : ndarray,
                  In sample selection, the matrix containing the selected targets, for use in fitting


    """

    def __init__(
        self,
        selection_type,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        self.selection_type = selection_type
        self.n_to_select = n_to_select
        self.score_threshold = score_threshold
        self.score_threshold_type = score_threshold_type
        self._first_score = None
        if self.score_threshold_type not in ["relative", "absolute"]:
            raise ValueError(
                "invalid score_threshold_type, expected one of 'relative' or 'absolute'"
            )

        self.full = full
        self.progress_bar = progress_bar
        self.random_state = random_state

    def fit(self, X, y=None, warm_start=False):
        """Learn the features to select.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training vectors.
        y : ndarray of shape (n_samples,), default=None
            Target values.
        warm_start : bool
            Whether the fit should continue after having already
            run, after increasing n_to_select.
            Assumes it is called with the same X and y

        Returns
        -------
        self : object
        """
        tags = self._get_tags()

        if self.selection_type == "feature":
            self._axis = 1
        elif self.selection_type == "sample":
            self._axis = 0
        else:
            raise ValueError("Only feature and sample selection supported.")

        if self.full and self.score_threshold is not None:
            raise ValueError(
                "You cannot specify both `score_threshold` and `full=True`."
            )

        if self.progress_bar is True:
            self.report_progress = get_progress_bar()
        elif self.progress_bar is False:
            self.report_progress = no_progress_bar

        params = dict(
            accept_sparse="csc",
            force_all_finite=not tags.get("allow_nan", True),
        )
        if self._axis == 1:
            params["ensure_min_features"] = 2
        else:
            params["ensure_min_samples"] = 2

        if y is not None:
            params["multi_output"] = True
            X, y = self._validate_data(X, y, **params)

            if len(y.shape) == 1:
                # force y to have multi_output 2D format even when it's 1D, since
                # many functions, most notably PCov routines, assume an array storage
                # format, most notably to compute (y @ y.T)
                y = y.reshape((len(y), 1))
        else:
            X = check_array(X, **params)

        n_to_select_from = X.shape[self._axis]

        error_msg = (
            "n_to_select must be either None, an "
            f"integer in [1, n_{self.selection_type}s] "
            "representing the absolute "
            f"number of {self.selection_type}s, or a float in (0, 1] "
            f"representing a percentage of {self.selection_type}s to "
            f"select. Got {self.n_to_select} {self.selection_type}s and "
            f"an input with {n_to_select_from} {self.selection_type}."
        )

        if self.n_to_select is None:
            n_iterations = n_to_select_from // 2
        elif isinstance(self.n_to_select, numbers.Integral):
            if not 0 < self.n_to_select <= n_to_select_from:
                raise ValueError(error_msg)
            n_iterations = self.n_to_select
        elif isinstance(self.n_to_select, numbers.Real):
            if not 0 < self.n_to_select <= 1:
                raise ValueError(error_msg)
            n_iterations = int(n_to_select_from * self.n_to_select)
        else:
            raise ValueError(error_msg)

        if warm_start:
            if not hasattr(self, "n_selected_") or getattr(self, "n_selected_") == 0:
                raise ValueError(
                    "Cannot fit with warm_start=True without having been previously initialized"
                )
            self._continue_greedy_search(X, y, n_iterations)
        else:
            self._init_greedy_search(X, y, n_iterations)

        n_iterations -= self.n_selected_

        for n in self.report_progress(range(n_iterations)):
            new_idx = self._get_best_new_selection(self.score, X, y)
            if new_idx is not None:
                self._update_post_selection(X, y, new_idx)
            else:
                warnings.warn(
                    f"Score threshold of {self.score_threshold} reached."
                    f"Terminating search at {self.n_selected_} / {self.n_to_select}."
                )
                self.X_selected_ = np.take(
                    self.X_selected_, np.arange(self.n_selected_), axis=self._axis
                )

                if hasattr(self, "y_selected_"):
                    self.y_selected_ = self.y_selected_[:n]

                self.selected_idx_ = self.selected_idx_[:n]
                self._postprocess(X, y)
                return self

        self._postprocess(X, y)
        return self

    def transform(self, X, y=None):
        """Reduce X to the selected features.

        Parameters
        ----------
        X : ndarray of shape [n_samples, n_features]
            The input samples.
        y : ignored

        Returns
        -------
        X_r : ndarray
            The selected subset of the input.
        """

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        mask = self.get_support()

        # note: we use _safe_tags instead of _get_tags because this is a
        # public Mixin.
        X = self._validate_data(
            X,
            dtype=None,
            accept_sparse="csr",
            force_all_finite=not _safe_tags(self, key="allow_nan"),
            reset=False,
            ensure_2d=self._axis,
        )

        if len(mask) != X.shape[self._axis]:
            raise ValueError("X has a different shape than during fitting.")

        if self._axis == 1:
            return X[:, safe_mask(X, mask)]
        else:
            return X[safe_mask(X, mask)]

    @abstractmethod
    def score(self, X, y):
        """
        A single str or a callable to evaluate the features or samples
        that is overwritten by the subclass.
        It is assumed that the next selection is that which maximizes the score.

        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.


        Parameters
        ----------
        X : ndarray of shape [n_samples, n_features]
            The input samples.
        y : ignored

        Returns
        -------
        score : ndarray of (n_to_select_from_)
            Scores of the given features or samples
        """
        pass

    def get_support(self, indices=False, ordered=False):
        """Get a mask, or integer index, of the subset

        Parameters
        ----------
        indices : bool, default=False
            If True, the return value will be an array of integers, rather
            than a bool mask.

        ordered : bool, default=False
            With indices, if True, the return value will be an array of integers, rather
            than a bool mask, in the order in which they were selected.

        Returns
        -------
        support : An index that selects the retained subset from a original vectors.
                  If indices is False, this is a bool array of shape [# input],
                  in which an element is True iff its corresponding feature or sample is selected
                  for retention. If indices is True, this is an integer array of shape
                  [# n_to_select] whose values are indices into the input vectors.

        """
        check_is_fitted(self, ["support_", "selected_idx_"])
        if indices:
            if ordered:
                return self.selected_idx_
            else:
                return list(sorted(self.selected_idx_))
        else:
            return self._get_support_mask()

    def _init_greedy_search(self, X, y, n_to_select):
        """Initializes the search. Prepares an array to store the selected features."""

        self.n_selected_ = 0

        sel_shape = list(X.shape)
        sel_shape[self._axis] = n_to_select

        self.X_selected_ = np.zeros(sel_shape, float)

        if y is not None and self._axis == 0:
            self.y_selected_ = np.zeros(
                (n_to_select, y.reshape(y.shape[0], -1).shape[1]), float
            )
        self.selected_idx_ = np.zeros((n_to_select), int)

    def _continue_greedy_search(self, X, y, n_to_select):
        """Continues the search. Prepares an array to store the selected features."""

        n_pad = [(0, 0), (0, 0)]
        n_pad[self._axis] = (0, n_to_select - self.n_selected_)

        self.X_selected_ = np.pad(
            self.X_selected_,
            n_pad,
            "constant",
            constant_values=0.0,
        )

        if hasattr(self, "y_selected_"):
            self.y_selected_ = np.pad(
                self.y_selected_,
                n_pad,
                "constant",
                constant_values=0.0,
            )

        old_idx = self.selected_idx_.copy()
        self.selected_idx_ = np.zeros((n_to_select), int)
        self.selected_idx_[: self.n_selected_] = old_idx

    def _get_best_new_selection(self, scorer, X, y):
        scores = scorer(X, y)

        max_score_idx = np.argmax(scores)
        if self.score_threshold is not None:
            if self._first_score is None:
                self._first_score = scores[max_score_idx]

            if self.score_threshold_type == "absolute":
                if scores[max_score_idx] < self.score_threshold:
                    return None

            if self.score_threshold_type == "relative":
                if scores[max_score_idx] / self._first_score < self.score_threshold:
                    return None

        return max_score_idx

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recently selected feature and increments the feature counter
        """

        if self._axis == 1:
            self.X_selected_[:, self.n_selected_] = np.take(
                X, last_selected, axis=self._axis
            )
        else:
            self.X_selected_[self.n_selected_] = np.take(
                X, last_selected, axis=self._axis
            )

            if hasattr(self, "y_selected_"):
                self.y_selected_[self.n_selected_] = y[last_selected]

        self.selected_idx_[self.n_selected_] = last_selected
        self.n_selected_ += 1

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which subset has been selected

        Raises
        ------
        NotFittedError
            If the selector has not yet been fitted

        Returns
        -------
        support : bool ndarray of shape [# input]
            An element is True iff its corresponding feature or sample is selected for
            retention.
        """
        check_is_fitted(self, ["support_"])
        return self.support_

    def _postprocess(self, X, y):
        """Post-process X and / or y when selection is finished"""
        self.support_ = np.full(X.shape[self._axis], False)
        self.support_[self.selected_idx_] = True

    def _more_tags(self):
        return {
            "requires_y": False,
        }


class _CUR(GreedySelector):
    """Transformer that performs Greedy Selection by choosing features
    which maximize the magnitude of the right or left singular vectors, consistent with
    classic CUR matrix decomposition.

    **WARNING**: This base class should never be directly instantiated.
    Instead, use :py:class:`skmatter.feature_selection.CUR` and
    :py:class:`skmatter.sample_selection.CUR`,
    which have the same constructor signature.

    Parameters
    ----------
    recompute_every : int
                      number of steps after which to recompute the pi score
                      defaults to 1, if 0 no re-computation is done

    k : int
        number of eigenvectors to compute the importance score with, defaults to 1

    tolerance: float
         threshold below which scores will be considered 0, defaults to 1E-12


    Attributes
    ----------

    X_current_ : ndarray (n_samples, n_features)
                  The original matrix orthogonalized by previous selections

    """

    def __init__(
        self,
        selection_type,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        self.k = k
        self.tolerance = tolerance
        self.recompute_every = recompute_every

        super().__init__(
            selection_type=selection_type,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )

    def score(self, X, y=None):
        r"""
        Returns the importance score of the given samples or features.

        NOTE: This function does not compute the importance score each time it
        is called, in order to avoid unnecessary computations. This is done
        by :py:func:`self._compute_pi`.

        Parameters
        ----------
        X : ndarray of shape [n_samples, n_features]
            The input samples.

        y : ignored

        Returns
        -------
        score : ndarray of (n_to_select_from_)
            :math:`\pi` importance for the given samples or features

        """

        return self.pi_

    def _init_greedy_search(self, X, y, n_to_select):
        """
        Initializes the search. Prepares an array to store the selected
        features and computes their initial importance score.
        """

        self.X_current_ = X.copy()
        self.pi_ = self._compute_pi(self.X_current_)

        super()._init_greedy_search(X, y, n_to_select)

    def _continue_greedy_search(self, X, y, n_to_select):
        """
        Continues the search. Prepares an array to store the selected
        features, orthogonalizes the features by those already selected,
        and computes their initial importance.
        """

        for c in self.selected_idx_:
            if self.recompute_every != 0 and (
                np.linalg.norm(np.take(self.X_current_, [c], axis=self._axis))
                > self.tolerance
            ):
                self._orthogonalize(last_selected=c)

        self.pi_ = self._compute_pi(self.X_current_)

        super()._continue_greedy_search(X, y, n_to_select)

    def _compute_pi(self, X, y=None):
        """
        For feature selection, the importance score :math:`\\pi` is the sum over
        the squares of the first :math:`k` components of the right singular vectors

        .. math::

            \\pi_j =
            \\sum_i^k \\left(\\mathbf{U}_\\mathbf{C}\\right)_{ij}^2.

        where :math:`\\mathbf{C} = \\mathbf{X}^T\\mathbf{X}`.

        For sample selection, the importance score :math:`\\pi` is the sum over
        the squares of the first :math:`k` components of the right singular vectors

        .. math::

            \\pi_j =
            \\sum_i^k \\left(\\mathbf{U}_\\mathbf{K}\\right)_{ij}^2.

        where :math:`\\mathbf{K} = \\mathbf{X}\\mathbf{X}^T`.

        Parameters
        ----------
        X : ndarray of shape [n_samples, n_features]
            The input samples.

        y : ignored

        Returns
        -------
        pi : ndarray of (n_to_select_from_)
            :math:`\\pi` importance for the given samples or features
        """

        if self._axis == 0:
            U, _, _ = scipy.sparse.linalg.svds(X, k=self.k, return_singular_vectors="u")
            U = np.real(U)
            new_pi = (U[:, : self.k] ** 2.0).sum(axis=1)
        else:
            _, _, Vt = scipy.sparse.linalg.svds(
                X, k=self.k, return_singular_vectors="vh"
            )
            new_pi = (np.real(Vt) ** 2.0).sum(axis=0)

        return new_pi

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recently selected feature, increments the feature counter,
        and, if the CUR is iterative (if recompute_every>0), orthogonalizes the remaining features by
        the most recently selected.
        """
        super()._update_post_selection(X, y, last_selected)

        if self.recompute_every != 0:
            self._orthogonalize(last_selected)

            if self.n_selected_ % self.recompute_every == 0:
                self.pi_ = self._compute_pi(self.X_current_)

        self.pi_[last_selected] = 0.0

    def _orthogonalize(self, last_selected):
        if self._axis == 1:
            self.X_current_ = X_orthogonalizer(
                x1=self.X_current_, c=last_selected, tol=self.tolerance
            )
        else:
            self.X_current_ = X_orthogonalizer(
                x1=self.X_current_.T, c=last_selected, tol=self.tolerance
            ).T


class _PCovCUR(GreedySelector):
    """Transformer that performs Greedy Selection by choosing features
    which maximize the magnitude of the right or left augmented singular vectors.
    This is done by employing the augmented kernel and covariance matrices,

    **WARNING**: This base class should never be directly instantiated.
    Instead, use :py:class:`skmatter.feature_selection.PCovCUR` and
    :py:class:`skmatter.sample_selection.PCovCUR`,
    which have the same constructor signature.

    Parameters
    ----------
    recompute_every : int
                      number of steps after which to recompute the pi score
                      defaults to 1, if 0 no re-computation is done

    k : int
        number of eigenvectors to compute the importance score with, defaults to 1

    tolerance: float
         threshold below which scores will be considered 0, defaults to 1E-12

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`. Stored in :py:attr:`self.mixing`.

    Attributes
    ----------

    X_current_ : ndarray (n_samples, n_features)
                  The original matrix orthogonalized by previous selections

    y_current_ : ndarray (n_samples, n_properties)
                The targets orthogonalized by a regression on
                the previous selections.

    """

    def __init__(
        self,
        selection_type,
        mixing=0.5,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        self.mixing = mixing

        self.k = k
        self.recompute_every = recompute_every
        self.tolerance = tolerance

        super().__init__(
            selection_type=selection_type,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )

    def score(self, X, y=None):
        """
        Returns the importance score of the given samples or features.

        NOTE: This function does not compute the importance score each time it
        is called, in order to avoid unnecessary computations. This is done
        by :py:func:`self._compute_pi`.

        Parameters
        ----------
        X : ignored

        y : ignored

        Returns
        -------
        score : ndarray of (n_to_select_from_)
            :math:`\\pi` importance for the given samples or features

        """

        return self.pi_

    def _init_greedy_search(self, X, y, n_to_select):
        """
        Initializes the search. Prepares an array to store the selected
        features and computes their initial importance score.
        """

        self.X_ref_ = X
        self.y_ref_ = y
        self.X_current_ = X.copy()
        if y is not None:
            self.y_current_ = y.copy()
        else:
            self.y_current_ = None
        self.pi_ = self._compute_pi(self.X_current_, self.y_current_)

        super()._init_greedy_search(X, y, n_to_select)

    def _continue_greedy_search(self, X, y, n_to_select):
        """
        Continues the search. Prepares an array to store the selected
        features, orthogonalizes the features by those already selected,
        and computes their initial importance.
        """

        for c in self.selected_idx_:
            if self.recompute_every != 0 and (
                np.linalg.norm(np.take(self.X_current_, [c], axis=self._axis))
                > self.tolerance
            ):
                self._orthogonalize(last_selected=c)

        self.pi_ = self._compute_pi(self.X_current_, self.y_current_)

        super()._continue_greedy_search(X, y, n_to_select)

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recently selected feature, increments the feature counter,
        and, if the CUR is iterative (if recompute_every>0), orthogonalizes the remaining features by
        the most recently selected.
        """
        super()._update_post_selection(X, y, last_selected)

        if self.recompute_every != 0:
            self._orthogonalize(last_selected)

            if self.n_selected_ % self.recompute_every == 0:
                self.pi_ = self._compute_pi(self.X_current_, self.y_current_)

        self.pi_[last_selected] = 0.0

    def _compute_pi(self, X, y=None):
        r"""
        For feature selection, the importance score :math:`\pi` is the sum over
        the squares of the first :math:`k` components of the right singular vectors

        .. math::

            \pi_j =
            \sum_i^k \left(\mathbf{U}_\mathbf{\tilde{C}}\right)_{ij}^2.

        where :math:`{\mathbf{\tilde{C}} = \alpha \mathbf{X}^T\mathbf{X} +
        (1 - \alpha)(\mathbf{X}^T\mathbf{X})^{-1/2}\mathbf{X}^T
        \mathbf{\hat{Y}\hat{Y}}^T\mathbf{X}(\mathbf{X}^T\mathbf{X})^{-1/2}}`
        for some mixing parameter :math:`{\alpha}`. When :math:`{\alpha = 1}`,
        this defaults to the covariance matrix
        :math:`{\mathbf{C} = \mathbf{X}^T\mathbf{X}}` used in CUR.

        For sample selection, the importance score :math:`\pi` is the sum over
        the squares of the first :math:`k` components of the right singular vectors

        .. math::

            \pi_j =
            \sum_i^k \left(\mathbf{U}_\mathbf{\tilde{K}}\right)_{ij}^2.

        where :math:`{\mathbf{\tilde{K}} = \alpha \mathbf{XX}^T +
        (1 - \alpha)\mathbf{\hat{Y}\hat{Y}}^T}` for some mixing parameter
        :math:`{\alpha}`. When :math:`{\alpha = 1}`, this defaults to the Gram
        matrix :math:`{\mathbf{K} = \mathbf{X}\mathbf{X}^T}`.

        Parameters
        ----------
        X : ndarray of shape [n_samples, n_features]
            The input samples.

        y : ignored

        Returns
        -------
        pi : ndarray of (n_to_select_from_)
            :math:`\pi` importance for the given samples or features
        """

        if self._axis == 0:
            pcovr_distance = pcovr_kernel(
                self.mixing,
                X,
                y,
            )
        else:
            pcovr_distance = pcovr_covariance(
                self.mixing,
                X,
                y,
                rcond=1e-12,
                rank=None,
            )

        if self.k < pcovr_distance.shape[0] - 1:
            v, U = eigsh(pcovr_distance, k=self.k, tol=1e-12)
        else:
            v, U = eigh(pcovr_distance)
        U = U[:, np.flip(np.argsort(v))]
        pi = (np.real(U)[:, : self.k] ** 2.0).sum(axis=1)

        return pi

    def _orthogonalize(self, last_selected):
        if self._axis == 1:
            self.X_current_ = X_orthogonalizer(
                x1=self.X_current_, c=last_selected, tol=self.tolerance
            )
        else:
            self.X_current_ = X_orthogonalizer(
                x1=self.X_current_.T, c=last_selected, tol=self.tolerance
            ).T
        if self.y_current_ is not None:
            if self._axis == 1:
                self.y_current_ = Y_feature_orthogonalizer(
                    self.y_current_, X=self.X_selected_, tol=self.tolerance
                )
            else:
                self.y_current_ = Y_sample_orthogonalizer(
                    self.y_ref_,
                    self.X_ref_,
                    y_ref=self.y_selected_[: self.n_selected_],
                    X_ref=self.X_selected_[: self.n_selected_],
                    tol=self.tolerance,
                )


class _FPS(GreedySelector):
    """
    Transformer that performs Greedy Selection using Farthest Point Sampling.

    **WARNING**: This base class should never be directly instantiated.
    Instead, use :py:class:`skmatter.feature_selection.FPS` and
    :py:class:`skmatter.sample_selection.FPS`,
    which have the same constructor signature.

    Parameters
    ----------

    initialize: int, list of int, or 'random', default=0
        Index of the first selection(s). If 'random', picks a random
        value when fit starts. Stored in :py:attr:`self.initialize`.


    """

    def __init__(
        self,
        selection_type,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        self.initialize = initialize

        super().__init__(
            selection_type=selection_type,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )

    def score(self, X, y=None):
        """
        Returns the Haussdorf distances of all samples to previous selections

        NOTE: This function does not compute the importance score each time it
        is called, in order to avoid unnecessary computations. The haussdorf
        distance is updated in :py:func:`self._update_haussdorf`

        Parameters
        ----------
        X : ignored
        y : ignored

        Returns
        -------
        haussdorf : Haussdorf distances
        """
        return self.haussdorf_

    def get_distance(self):
        """

        Traditional FPS employs a column-wise Euclidean
        distance for feature selection, which can be expressed using the covariance matrix
        :math:`\\mathbf{C} = \\mathbf{X} ^ T \\mathbf{X}`

        .. math::
            \\operatorname{d}_c(i, j) = C_{ii} - 2 C_{ij} + C_{jj}.

        For sample selection, this is a row-wise Euclidean distance, which can
        be expressed in terms of the Gram matrix :math:`\\mathbf{K} = \\mathbf{X} \\mathbf{X} ^ T`

        .. math::
            \\operatorname{d}_r(i, j) = K_{ii} - 2 K_{ij} + K_{jj}.

        Returns
        -------

        haussdorf : ndarray of shape (`n_to_select_from_`)
                     the minimum distance from each point to the set of selected
                     points. once a point is selected, the distance is not updated;
                     the final list will reflect the distances when selected.

        """
        return self.haussdorf_

    def get_select_distance(self):
        """

        Returns
        -------

        haussdorf_at_select : ndarray of shape (`n_to_select`)
                     at the time of selection, the minimum distance from each
                     selected point to the set of previously selected points.

        """
        mask = self.get_support(indices=True, ordered=True)
        return self.haussdorf_at_select_[mask]

    def _init_greedy_search(self, X, y, n_to_select):
        """
        Initializes the search. Prepares an array to store the selections,
        makes the initial selection (unless provided), and
        computes the starting haussdorf distances.
        """

        super()._init_greedy_search(X, y, n_to_select)

        self.norms_ = (X**2).sum(axis=abs(self._axis - 1))
        self.haussdorf_ = np.full(X.shape[self._axis], np.inf)
        self.haussdorf_at_select_ = np.full(X.shape[self._axis], np.inf)

        if self.initialize == "random":
            random_state = check_random_state(self.random_state)
            initialize = random_state.randint(X.shape[self._axis])
            self.selected_idx_[0] = initialize
            self._update_post_selection(X, y, self.selected_idx_[0])
        elif isinstance(self.initialize, numbers.Integral):
            initialize = self.initialize
            self.selected_idx_[0] = initialize
            self._update_post_selection(X, y, self.selected_idx_[0])
        elif isinstance(self.initialize, list) and all(
            [isinstance(i, numbers.Integral) for i in self.initialize]
        ):
            for i, val in enumerate(self.initialize):
                self.selected_idx_[i] = val
                self._update_post_selection(X, y, self.selected_idx_[i])
        else:
            raise ValueError("Invalid value of the initialize parameter")

    def _update_haussdorf(self, X, y, last_selected):
        self.haussdorf_at_select_[last_selected] = self.haussdorf_[last_selected]

        # distances of all points to the new point
        if self._axis == 1:
            new_dist = (
                self.norms_ + self.norms_[last_selected] - 2 * X[:, last_selected].T @ X
            )
        else:
            new_dist = (
                self.norms_ + self.norms_[last_selected] - 2 * X[last_selected] @ X.T
            )

        # update in-place the Haussdorf distance list
        np.minimum(self.haussdorf_, new_dist, self.haussdorf_)

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recent selections, increments the counter,
        and, recomputes haussdorf distances.
        """
        self._update_haussdorf(X, y, last_selected)
        super()._update_post_selection(X, y, last_selected)


class _PCovFPS(GreedySelector):
    """
    Transformer that performs Greedy Selection using PCovR-weighted
    Farthest Point Sampling.
    In PCov-FPS, a modified covariance or Gram matrix
    is used to express the distances.

    For sample selection, this is a modified kernel matrix.

    Parameters
    ----------

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`

    initialize: int or 'random', default=0
        Index of the first selection. If 'random', picks a random
        value when fit starts.

    """

    def __init__(
        self,
        selection_type,
        mixing=0.5,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        if mixing == 1.0:
            raise ValueError(
                "Mixing = 1.0 corresponds to traditional FPS."
                "Please use the FPS class."
            )

        self.mixing = mixing
        self.initialize = initialize

        super().__init__(
            selection_type=selection_type,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )

    def score(self, X, y=None):
        """
        Returns the Haussdorf distances of all samples to previous selections

        NOTE: This function does not compute the importance score each time it
        is called, in order to avoid unnecessary computations. The haussdorf
        distance is updated in :py:func:`self._update_haussdorf`

        Parameters
        ----------
        X : ignored
        y : ignored

        Returns
        -------
        haussdorf : Haussdorf distances
        """
        return self.haussdorf_

    def get_distance(self):
        """

        Returns
        -------

        haussdorf : ndarray of shape (`n_to_select_from_`)
                     the minimum distance from each point to the set of selected
                     points. once a point is selected, the distance is not updated;
                     the final list will reflect the distances when selected.

        """
        return self.haussdorf_

    def get_select_distance(self):
        """

        Returns
        -------

        haussdorf_at_select : ndarray of shape (`n_to_select`)
                     at the time of selection, the minimum distance from each
                     selected point to the set of previously selected points.

        """
        mask = self.get_support(indices=True, ordered=True)
        return self.haussdorf_at_select_[mask]

    def _init_greedy_search(self, X, y, n_to_select):
        """
        Initializes the search. Prepares an array to store the selections,
        makes the initial selection (unless provided), and
        computes the starting haussdorf distances.
        """

        super()._init_greedy_search(X, y, n_to_select)

        if self._axis == 1:
            self.pcovr_distance_ = pcovr_covariance(mixing=self.mixing, X=X, Y=y)
        else:
            self.pcovr_distance_ = pcovr_kernel(mixing=self.mixing, X=X, Y=y)

        self.norms_ = np.diag(self.pcovr_distance_)

        if self.initialize == "random":
            random_state = check_random_state(self.random_state)
            initialize = random_state.randint(X.shape[self._axis])
        elif isinstance(self.initialize, numbers.Integral):
            initialize = self.initialize
        else:
            raise ValueError("Invalid value of the initialize parameter")

        self.selected_idx_[0] = initialize
        self.haussdorf_ = np.full(X.shape[self._axis], np.inf)
        self.haussdorf_at_select_ = np.full(X.shape[self._axis], np.inf)
        self._update_post_selection(X, y, self.selected_idx_[0])

    def _update_haussdorf(self, X, y, last_selected):
        self.haussdorf_at_select_[last_selected] = self.haussdorf_[last_selected]

        # distances of all points to the new point
        new_dist = (
            self.norms_
            + self.norms_[last_selected]
            - 2 * np.take(self.pcovr_distance_, last_selected, axis=self._axis)
        )

        # update in-place the Haussdorf distance list
        np.minimum(self.haussdorf_, new_dist, self.haussdorf_)

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recent selections, increments the counter,
        and, recomputes haussdorf distances.
        """
        self._update_haussdorf(X, y, last_selected)
        super()._update_post_selection(X, y, last_selected)

    def _more_tags(self):
        """
        Pass that this method requires a target vector
        """
        return {
            "requires_y": True,
        }



import unittest

import numpy as np
import sklearn
from sklearn.preprocessing import StandardScaler

from skmatter.preprocessing import StandardFlexibleScaler


class ScalerTests(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

    def test_sample_weights(self):
        """Checks that sample weights of one are equal to the unweighted case and that the nonuniform weights are different from the unweighted case"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        equal_wts = np.ones(len(X))
        nonequal_wts = self.random_state.uniform(0, 100, size=(len(X),))
        model = StandardFlexibleScaler()
        weighted_model = StandardFlexibleScaler()
        X_unweighted = model.fit_transform(X)
        X_equal_weighted = weighted_model.fit_transform(X, sample_weight=equal_wts)
        self.assertTrue((np.isclose(X_unweighted, X_equal_weighted, atol=1e-12)).all())
        X_nonequal_weighted = weighted_model.fit_transform(
            X, sample_weight=nonequal_wts
        )
        self.assertFalse(
            (np.isclose(X_unweighted, X_nonequal_weighted, atol=1e-12)).all()
        )

    def test_invalid_sample_weights(self):
        """Checks that weights must be 1D array with the same length as the number of samples"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        wts_len = np.ones(len(X) + 1)
        wts_dim = np.ones((len(X), 2))
        model = StandardFlexibleScaler()
        with self.assertRaises(ValueError):
            model.fit_transform(X, sample_weight=wts_len)
        with self.assertRaises(ValueError):
            model.fit_transform(X, sample_weight=wts_dim)

    def test_fit_transform_pf(self):
        """Checks that in the case of normalization by columns,
        the result is the same as in the case of using the package from sklearn
        """
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=True)
        transformed_skmatter = model.fit_transform(X)
        transformed_sklearn = StandardScaler().fit_transform(X)
        self.assertTrue(
            (np.isclose(transformed_sklearn, transformed_skmatter, atol=1e-12)).all()
        )

    def test_fit_transform_npf(self):
        """Checks that the entire matrix is correctly normalized
        (not column-wise). Compare with the value calculated
        directly from the equation.
        """
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=False)
        X_tr = model.fit_transform(X)
        mean = X.mean(axis=0)
        var = ((X - mean) ** 2).mean(axis=0)
        scale = np.sqrt(var.sum())
        X_ex = (X - mean) / scale
        self.assertTrue((np.isclose(X_ex, X_tr, atol=1e-12)).all())

    def test_transform(self):
        """Checks the transformation relative
        to the reference matrix.
        """
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=True)
        model.fit(X)
        Y = self.random_state.uniform(0, 100, size=(3, 3))
        Y_tr = model.transform(Y)
        mean = X.mean(axis=0)
        var = ((X - mean) ** 2).mean(axis=0)
        scale = np.sqrt(var)
        Y_ex = (Y - mean) / scale
        self.assertTrue((np.isclose(Y_tr, Y_ex, atol=1e-12)).all())

    def test_inverse_transform(self):
        """Checks the inverse transformation with
        respect to the reference matrix.
        """
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=True)
        model.fit(X)
        Y = self.random_state.uniform(0, 100, size=(3, 3))
        Y_tr = model.transform(Y)
        Y = np.around(Y, decimals=4)
        Y_inv = np.around((model.inverse_transform(Y_tr)), decimals=4)
        self.assertTrue((np.isclose(Y, Y_inv, atol=1e-12)).all())
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=False)
        model.fit(X)
        Y = self.random_state.uniform(0, 100, size=(3, 3))
        Y_tr = model.transform(Y)
        Y = np.around(Y, decimals=4)
        Y_inv = np.around((model.inverse_transform(Y_tr)), decimals=4)
        self.assertTrue((np.isclose(Y, Y_inv, atol=1e-12)).all())

    def test_NotFittedError_transform(self):
        """Checks that an error is returned when
        trying to use the transform function
        before the fit function"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler(column_wise=True)
        with self.assertRaises(sklearn.exceptions.NotFittedError):
            model.transform(X)

    def test_shape_inconsistent_transform(self):
        """Checks that an error is returned when attempting
        to use the transform function with mismatched matrix sizes."""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        X_test = self.random_state.uniform(0, 100, size=(4, 4))
        model = StandardFlexibleScaler(column_wise=True)
        model.fit(X)
        with self.assertRaises(ValueError):
            model.transform(X_test)

    def test_shape_inconsistent_inverse(self):
        """Checks that an error is returned when attempting
        to use the inverse transform function with mismatched matrix sizes."""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        X_test = self.random_state.uniform(0, 100, size=(4, 4))
        model = StandardFlexibleScaler(column_wise=True)
        model.fit(X)
        with self.assertRaises(ValueError):
            model.inverse_transform(X_test)

    def test_NotFittedError_inverse(self):
        """Checks that an error is returned when
        trying to use the inverse transform function
        before the fit function"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        model = StandardFlexibleScaler()
        with self.assertRaises(sklearn.exceptions.NotFittedError):
            model.inverse_transform(X)

    def test_ValueError_column_wise(self):
        """Checks that the matrix cannot be normalized
        across columns if there is a zero variation column."""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        X[0][0] = X[1][0] = X[2][0] = 2
        model = StandardFlexibleScaler(column_wise=True)
        with self.assertRaises(ValueError):
            model.fit(X)

    def test_atol(self):
        """Checks that we can define absolute tolerance and it control the
        minimal variance of columns ot the whole matrix"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        atol = ((X[:, 0] - X[:, 0].mean(axis=0)) ** 2).mean(axis=0) + 1e-8
        model = StandardFlexibleScaler(column_wise=True, atol=atol, rtol=0)
        with self.assertRaises(ValueError):
            model.fit(X)
        atol = (X - X.mean(axis=0) ** 2).mean(axis=0) + 1e-8
        model = StandardFlexibleScaler(column_wise=False, atol=atol, rtol=0)
        with self.assertRaises(ValueError):
            model.fit(X)

    def test_rtol(self):
        """Checks that we can define relative tolerance and it control the
        minimal variance of columns or the whole matrix"""
        X = self.random_state.uniform(0, 100, size=(3, 3))
        mean = X[:, 0].mean(axis=0)
        rtol = ((X[:, 0] - mean) ** 2).mean(axis=0) / mean + 1e-8
        model = StandardFlexibleScaler(column_wise=True, atol=0, rtol=rtol)
        with self.assertRaises(ValueError):
            model.fit(X)
        mean = X.mean(axis=0)
        rtol = ((X - mean) ** 2).mean(axis=0) / mean + 1e-8
        model = StandardFlexibleScaler(column_wise=False, atol=0, rtol=rtol)
        with self.assertRaises(ValueError):
            model.fit(X)

    def test_ValueError_full(self):
        """Checks that the matrix cannot be normalized
        if there is a zero variation matrix."""
        X = np.array([2, 2, 2]).reshape(-1, 1)
        model = StandardFlexibleScaler(column_wise=False)
        with self.assertRaises(ValueError):
            model.fit(X)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from sklearn.datasets import load_diabetes as get_dataset

from skmatter.feature_selection import PCovCUR


class TestPCovCUR(unittest.TestCase):
    def setUp(self):
        self.X, self.y = get_dataset(return_X_y=True)
        self.idx = [2, 8, 3, 4, 1, 7, 5, 9, 6]

    def test_known(self):
        """
        This test checks that the model returns a known set of indices
        """

        selector = PCovCUR(n_to_select=9)
        selector.fit(self.X, self.y)

        self.assertTrue(np.allclose(selector.selected_idx_, self.idx))

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new instance
        """

        selector = PCovCUR(n_to_select=1)
        selector.fit(self.X, self.y)

        for i in range(len(self.idx) - 2):
            selector.n_to_select += 1
            selector.fit(self.X, warm_start=True)
            self.assertEqual(selector.selected_idx_[i], self.idx[i])

    def test_non_it(self):
        """
        This test checks that the model can be run non-iteratively
        """
        self.idx = [2, 8, 3, 6, 7, 9, 1, 0, 5]
        selector = PCovCUR(n_to_select=9, recompute_every=0)
        selector.fit(self.X, self.y)

        self.assertTrue(np.allclose(selector.selected_idx_, self.idx))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import numbers
from time import time

import numpy as np
from sklearn.utils import check_random_state

from .._selection import GreedySelector


class VoronoiFPS(GreedySelector):
    """
    In FPS, points are selected based upon their Hausdorff distance to
    previous selections, i.e. the minimum distance between a given point and
    any previously selected points. This implicitly constructs a Voronoi
    tessellation which is updated with each new selection, as each unselected
    point "belongs" to the Voronoi polyhedron of the nearest previous selection.

    This implicit tessellation enabled a more efficient evaluation of the FPS --
    at each iteration, we need only consider for selection those points at the
    boundaries of the Voronoi polyhedra, and when updating the tessellation
    we need only consider moving those points whose Hausdorff distance is
    greater than half of the distance between the corresponding Voronoi center
    and the newly selected point, per the triangle equality.

    .. image:: VoronoiFPS-Schematic.pdf

    To demonstrate the algorithm behind Voronoi FPS, let :math:`*_{m+1}` be a new chosen point,
    :math:`v(j)` was chosen earlier, :math:`j` is a point in the polyhedron with center
    :math:`v(j)`. From the inequalities of the triangle one can easily see that if
    :math:`d(v(j),j)<d(*_{m+1}, j)/2`, then point :math:`j` is guaranteed not to fall
    into the formed polyhedron and the distance to it can be uncalculated.

    This algorithm is particularly appealing when using a non-Euclidean or
    computationally-intensive distance metric, for which the
    decrease in computational time due to the reduction in distance
    calculations outweighs the increase from book-keeping. For simple
    metrics (such as Euclidean distance), VoronoiFPS will likely not
    accelerate, and may decelerate, computations when compared
    to FPS.

    Parameters
    ----------

    n_trial_calculation: integer, default=4
        Number of calculations used for the switching point between Voronoi FPS
        and traditional FPS (for detail look at full_fraction).

    full_fraction: float, default=None
        Proportion of calculated distances from the total number of features at
        which the switch from Voronoi FPS to FPS occurs.
        At a certain number of distances to be calculated,
        the use of Voronoi FPS becomes unreasonably expensive due to the associated
        costs connected with reading data from the memory. The switching point
        depends on many conditions, and it is determined "in situ" for optimal
        use of the algorithm. Determination is done with a few test calculations
        and memory operations.
    """

    def __init__(
        self, n_trial_calculation=4, full_fraction=None, initialize=0, **kwargs
    ):
        self.n_trial_calculation = n_trial_calculation
        self.full_fraction = full_fraction
        self.initialize = initialize
        super().__init__(selection_type="sample", **kwargs)

    def score(self, X=None, y=None):
        """
        Returns the Haussdorf distances of all samples to previous selections

        NOTE: This function does not compute the importance score each time it
        is called, in order to avoid unnecessary computations. The haussdorf
        distance is updated in :py:func:`self._update_post_selection`

        Parameters
        ----------
        X : ignored
        y : ignored

        Returns
        -------
        haussdorf : Haussdorf distances
        """
        return self.haussdorf_

    def get_distance(self):
        """

        Traditional FPS employs a column-wise Euclidean
        distance for feature selection, which can be expressed using the covariance matrix
        :math:`\\mathbf{C} = \\mathbf{X} ^ T \\mathbf{X}`

        .. math::
            \\operatorname{d}_c(i, j) = C_{ii} - 2 C_{ij} + C_{jj}.

        For sample selection, this is a row-wise Euclidean distance, which can
        be expressed in terms of the Gram matrix :math:`\\mathbf{K} = \\mathbf{X} \\mathbf{X} ^ T`

        .. math::
            \\operatorname{d}_r(i, j) = K_{ii} - 2 K_{ij} + K_{jj}.

        Returns
        -------

        haussdorf : ndarray of shape (`n_to_select_from_`)
                     the minimum distance from each point to the set of selected
                     points. once a point is selected, the distance is not updated;
                     the final list will reflect the distances when selected.

        """
        return self.haussdorf_

    def get_select_distance(self):
        """

        Returns
        -------

        haussdorf_at_select : ndarray of shape (`n_to_select`)
                     at the time of selection, the minimum distance from each
                     selected point to the set of previously selected points.

        """
        mask = self.get_support(indices=True, ordered=True)
        return self.haussdorf_at_select_[mask]

    def _init_greedy_search(self, X, y, n_to_select):
        """
        Initializes the search. Prepares an array to store the selected
        features. This function also determines the switching point if it was not
        given during initialization.
        The point is that when calculating distances in Voronoi_FPS it is
        necessary to "jump" between indexes in the array. This results in a
        significant increase in time. Therefore, if you need to recalculate a
        large number of distances, it is more advantageous to run simple
        calculation along the whole matrix.
        """

        n_to_select_from = X.shape[0]
        self.vlocation_of_idx = np.full(n_to_select_from, 1)
        # index of the voronoi cell associated with each of the columns of X

        self.dSL_ = np.zeros(self.n_to_select, float)
        # distance between new selected point and previously
        # selected points

        if self.full_fraction is None:
            simple_fps_timing = -time()
            if not isinstance(self.n_trial_calculation, numbers.Integral):
                raise TypeError("Number of trial calculation should be integer")
            if self.n_trial_calculation == 0:
                raise ValueError(
                    "Number of trial calculation should be more or equal to 1"
                )
            for i in range(self.n_trial_calculation):
                _ = X @ X[[0]].T
            simple_fps_timing += time()
            simple_fps_timing /= self.n_trial_calculation

            lower_fraction = 0
            top_fraction = 1
            while top_fraction - lower_fraction > 0.01:
                voronoi_fps_times = np.zeros(self.n_trial_calculation)
                self.full_fraction = (top_fraction + lower_fraction) / 2
                for i in range(self.n_trial_calculation):
                    random_state = check_random_state(self.random_state)
                    sel = random_state.randint(
                        n_to_select_from,
                        size=int(n_to_select_from * self.full_fraction),
                    )
                    voronoi_fps_times[i] = -time()
                    _ = X[0] @ X[sel].T
                    voronoi_fps_times[i] += time()
                voronoi_fps_timing = np.sum(voronoi_fps_times)
                voronoi_fps_timing /= self.n_trial_calculation
                if voronoi_fps_timing < simple_fps_timing:
                    lower_fraction = self.full_fraction
                else:
                    top_fraction = self.full_fraction

            self.full_fraction = lower_fraction
        else:
            if isinstance(self.full_fraction, numbers.Real):
                if not 0 < self.full_fraction <= 1:
                    raise ValueError(
                        "Switching point should be real and more than 0 and less than 1. "
                        f"Received {self.full_fraction}"
                    )
            else:
                raise ValueError(
                    "Switching point should be real and more than 0 and less than 1. "
                    f"Received {self.full_fraction}"
                )

        super()._init_greedy_search(X, y, n_to_select)

        self.norms_ = (X**2).sum(axis=abs(self._axis - 1))

        if self.initialize == "random":
            random_state = check_random_state(self.random_state)
            initialize = random_state.randint(X.shape[self._axis])
        elif isinstance(self.initialize, numbers.Integral):
            initialize = self.initialize
        else:
            raise ValueError("Invalid value of the initialize parameter")

        self.selected_idx_[0] = initialize
        self.haussdorf_ = np.full(X.shape[self._axis], np.inf)
        self.haussdorf_at_select_ = np.full(X.shape[self._axis], np.inf)
        self._update_post_selection(X, y, self.selected_idx_[0])

    def _continue_greedy_search(self, X, y, n_to_select):
        """Continues the search. Prepares an array to store the selected
        features."""

        super()._continue_greedy_search(X, y, n_to_select)

        n_pad = n_to_select - self.n_selected_

        self.dSL_ = np.pad(self.dSL_, (0, n_pad), "constant", constant_values=0)

    def _get_active(self, X, last_selected):
        """
        Finds the indices of the Voronoi cells that might change due to the fact
        that a new point has been selected (index is last_selected,
        because we haven't called super() yet, self.n_selected_ has not been
        incremented yet)

        This function calculates the distances between the last selected point
        and the previously selected points. Next, the list of points to which
        the distance is to be calculated is shortened.

        These are the points for which need to recompute distances.
        Let:
        L is the last point selected;
        S are the selected points from before this iteration;
        X are the candidates;
        The logic here is that we want to check if d(XL) can be smaller than
        min(d(X,S)) (which is stored in self.haussdorf_)
        now, if a point belongs to the Voronoi cell of S then
        min(d(X,S_i))=d(X,S). Triangle inequality implies that
        d(S,L) < |d(X,S) + d(L,X)| so we just need to check if
        |d(X,S) - d(S,L)|>= d(X,S) to know that we don't need to check X.
        but |d(X,S) - d(S,L)|^2>= d(X,S)^2 if and only if d(S,L)/2 > d(S,X)
        """

        if not hasattr(self, "n_selected_") or self.n_selected_ == 0:
            return np.arange(X.shape[0], dtype=int)

        else:
            self.dSL_[: self.n_selected_] = (
                self.norms_[self.selected_idx_[: self.n_selected_]]
                + self.norms_[last_selected]
                - 2 * (self.X_selected_[: self.n_selected_] @ X[last_selected].T)
            ) * 0.25
            # calculation in a single block

            active_points = np.where(
                self.dSL_[self.vlocation_of_idx] < self.haussdorf_
            )[0]

            return active_points

    def _update_post_selection(self, X, y, last_selected):
        """
        Saves the most recently selected feature, increments the feature counter
        and update the haussdorf distances
        Let:
        L is the last point selected;
        S are the selected points from before this iteration;
        X is the one active point;
        This function calculates d(L, X) and checks the condition
        d(L, X)< min d(X, S_i). If so, we move X to a new polyhedron.
        If the number of active points is too high, it is faster to calculate
        the distances between L and all the points in the dataset.
        """

        self.haussdorf_at_select_[last_selected] = self.haussdorf_[last_selected]
        active_points = self._get_active(X, last_selected)

        if len(active_points) > 0:
            if len(active_points) / X.shape[0] > self.full_fraction:
                self.new_dist_ = (
                    self.norms_
                    + self.norms_[last_selected]
                    - 2 * X[last_selected] @ X.T
                )
            else:
                self.new_dist_ = self.haussdorf_.copy()

                self.new_dist_[active_points] = (
                    self.norms_[active_points]
                    + self.norms_[last_selected]
                    - 2 * X[last_selected] @ X[active_points].T
                )
                self.new_dist_[last_selected] = 0

            updated_points = np.where(self.new_dist_ < self.haussdorf_)[0]
            np.minimum(
                self.haussdorf_, self.new_dist_, self.haussdorf_, casting="unsafe"
            )
        else:
            updated_points = np.array([])

        if len(updated_points) > 0:
            self.vlocation_of_idx[updated_points] = self.n_selected_

        self.vlocation_of_idx[last_selected] = self.n_selected_
        super()._update_post_selection(X, y, last_selected)



import unittest

from sklearn.datasets import load_diabetes as get_dataset

from skmatter.feature_selection import PCovFPS


class TestPCovFPS(unittest.TestCase):
    def setUp(self):
        self.X, self.y = get_dataset(return_X_y=True)
        self.idx = [0, 2, 6, 7, 1, 3, 4]

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new number of
        features and `warm_start`
        """

        selector = PCovFPS(n_to_select=1, initialize=self.idx[0])
        selector.fit(self.X, y=self.y)

        for i in range(2, len(self.idx)):
            selector.n_to_select = i
            selector.fit(self.X, y=self.y, warm_start=True)
            self.assertEqual(selector.selected_idx_[i - 1], self.idx[i - 1])

    def test_no_mixing_1(self):
        """
        This test checks that the model throws an error when mixing = 1.0
        """

        with self.assertRaises(ValueError) as cm:
            _ = PCovFPS(n_to_select=1, mixing=1.0)
        self.assertEqual(
            str(cm.exception),
            "Mixing = 1.0 corresponds to traditional FPS." "Please use the FPS class.",
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
from sklearn import exceptions

from sklearn.datasets import fetch_california_housing as load
from skmatter.sample_selection import CUR


class TestCUR(unittest.TestCase):
    def setUp(self):
        self.X, _ = load(return_X_y=True)
        self.X = self.X[:1000]
        self.n_select = min(20, min(self.X.shape) // 2)

    def test_bad_transform(self):
        selector = CUR(n_to_select=2)
        with self.assertRaises(exceptions.NotFittedError):
            _ = selector.transform(self.X)

    def test_restart(self):
        """
        This test checks that the model can be restarted with a new instance
        """

        ref_selector = CUR(n_to_select=self.n_select)
        ref_idx = ref_selector.fit(self.X).selected_idx_

        selector = CUR(n_to_select=1)
        selector.fit(self.X)

        for i in range(len(ref_idx) - 2):
            selector.n_to_select += 1
            selector.fit(self.X, warm_start=True)
            self.assertEqual(selector.selected_idx_[i], ref_idx[i])

    def test_non_it(self):
        """
        This test checks that the model can be run non-iteratively
        """

        K = self.X @ self.X.T
        _, UK = np.linalg.eigh(K)
        ref_idx = np.argsort(-(UK[:, -1] ** 2.0))[: self.n_select]

        selector = CUR(n_to_select=len(ref_idx), recompute_every=0)
        selector.fit(self.X)

        self.assertTrue(np.allclose(selector.selected_idx_, ref_idx))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

import numpy as np
import sklearn

from skmatter.preprocessing import SparseKernelCenterer


class SparseKernelTests(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = np.random.RandomState(0)

    def test_sample_weights(self):
        """Checks that sample weights of one are equal to the unweighted case and that the nonuniform weights are different from the unweighted case"""
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        equal_wts = np.ones(len(Knm))
        nonequal_wts = self.random_state.uniform(-1, 1, size=(len(Knm),))
        model = SparseKernelCenterer()
        weighted_model = SparseKernelCenterer()
        Knm_unweighted = model.fit_transform(Knm, Kmm)
        Knm_equal_weighted = weighted_model.fit_transform(
            Knm, Kmm, sample_weight=equal_wts
        )
        Knm_nonequal_weighted = weighted_model.fit_transform(
            Knm, Kmm, sample_weight=nonequal_wts
        )
        self.assertTrue(
            (np.isclose(Knm_unweighted, Knm_equal_weighted, atol=1e-12)).all()
        )
        self.assertFalse(
            (np.isclose(Knm_unweighted, Knm_nonequal_weighted, atol=1e-12)).all()
        )

    def test_invalid_sample_weights(self):
        """Checks that weights must be 1D array with the same length as the number of samples"""
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        wts_len = np.ones(len(Knm) + 1)
        wts_dim = np.ones((len(Knm), 2))
        model = SparseKernelCenterer()
        with self.assertRaises(ValueError):
            model.fit_transform(Knm, Kmm, sample_weight=wts_len)
        with self.assertRaises(ValueError):
            model.fit_transform(Knm, Kmm, sample_weight=wts_dim)

    def test_Square_Kmm(self):
        """Checks that the passed active kernel is square"""

        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X.T

        model = SparseKernelCenterer()
        with self.assertRaises(ValueError) as cm:
            model.fit(Knm, Kmm)
        self.assertEqual(str(cm.exception), "The active kernel is not square.")

    def test_LatterDim(self):
        """Checks that a matrix must have the same latter dimension as its active counterpart cannot be normalized."""

        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X.T
        Kmm = X_sparse @ X_sparse.T

        model = SparseKernelCenterer()
        with self.assertRaises(ValueError) as cm:
            model.fit(Knm, Kmm)
        self.assertEqual(
            str(cm.exception),
            "The reference kernel is not " "commensurate shape with the active kernel.",
        )

    def test_new_kernel(self):
        """Checks that it is impossible to normalize
        a matrix with a non-coincident size with the reference."""
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        Knm2 = X @ X.T
        model = SparseKernelCenterer()
        model = model.fit(Knm, Kmm)
        with self.assertRaises(ValueError) as cm:
            model.transform(Knm2)
        self.assertEquals(
            str(cm.exception),
            "The reference kernel and received kernel have different shape",
        )

    def test_NotFittedError_transform(self):
        """Checks that an error is returned when
        trying to use the transform function
        before the fit function"""
        K = self.random_state.uniform(0, 100, size=(3, 3))
        model = SparseKernelCenterer()
        with self.assertRaises(sklearn.exceptions.NotFittedError):
            model.transform(K)

    def test_fit_transform(self):
        """Checks that the kernel is correctly normalized.
        Compare with the value calculated
        directly from the equation.
        """

        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        model = SparseKernelCenterer(rcond=1e-12)
        Ktr = model.fit_transform(Knm, Kmm)

        Knm_mean = Knm.mean(axis=0)

        Kc = Knm - Knm_mean

        Khat = Kc @ np.linalg.pinv(Kmm, rcond=1e-12) @ Kc.T

        Kc /= np.sqrt(np.trace(Khat) / Khat.shape[0])

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_center_only(self):
        """Checks that the kernel is correctly centered, but not normalized.
        Compare with the value calculated
        directly from the equation.
        """
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        model = SparseKernelCenterer(with_center=True, with_trace=False, rcond=1e-12)
        Ktr = model.fit_transform(Knm, Kmm)

        Knm_mean = Knm.mean(axis=0)

        Kc = Knm - Knm_mean

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_trace_only(self):
        """Checks that the kernel is correctly normalized, but not centered.
        Compare with the value calculated
        directly from the equation.
        """
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        model = SparseKernelCenterer(with_center=False, with_trace=True, rcond=1e-12)
        Ktr = model.fit_transform(Knm, Kmm)

        Kc = Knm.copy()

        Khat = Kc @ np.linalg.pinv(Kmm, rcond=1e-12) @ Kc.T

        Kc /= np.sqrt(np.trace(Khat) / Khat.shape[0])

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())

    def test_no_preprocessing(self):
        """Checks that the kernel is unchanged
        if no preprocessing is specified.
        """
        X = self.random_state.uniform(-1, 1, size=(4, 5))
        X_sparse = self.random_state.uniform(-1, 1, size=(3, 5))

        Knm = X @ X_sparse.T
        Kmm = X_sparse @ X_sparse.T

        model = SparseKernelCenterer(with_center=False, with_trace=False, rcond=1e-12)
        Ktr = model.fit_transform(Knm, Kmm)

        Kc = Knm.copy()

        self.assertTrue((np.isclose(Ktr, Kc, atol=1e-12)).all())


if __name__ == "__main__":
    unittest.main()



import numpy as np
from sklearn.base import (
    BaseEstimator,
    TransformerMixin,
)
from sklearn.preprocessing._data import KernelCenterer
from sklearn.utils.validation import (
    FLOAT_DTYPES,
    _check_sample_weight,
    check_is_fitted,
)


class StandardFlexibleScaler(TransformerMixin, BaseEstimator):
    """Standardize features by removing the mean and scaling to unit variance.
    Reduce the mean of the column to zero and, in the case of `column_wise=True`
    the variance of each column equal to one / number of columns.
    The standard score of a sample `x` is calculated as:

        z = (x - u) / s

    where `u` is the mean of the samples if `with_mean`, otherwise zero,
    and `s` is the standard deviation of the samples if `with_std` or one.

    Centering and scaling can occur independently for each feature by calculating
    the appropriate statistics for the input or for the
    whole matrix (`column_wise=False`). The mean and standard deviation are then
    stored for use on later data using :py:meth:`transform`.

    Standardization of a dataset is a common requirement for many
    machine learning estimators: an improperly scaled / centered
    dataset may result in anomalous behavior.

    At the same time, depending on the conditions of the task, it may be necessary
    to preserve the ratio in the scale between the features (for example, in the
    case where the feature matrix is something like a covariance matrix), so
    the standardization should be carried out for the whole matrix,
    as opposed to the individual columns,
    as is done in `sklearn.preprocessing.StandardScaler`.

    Parameters
    ----------
    with_mean: bool, default=True
        If True, center the data before scaling. If False, keep the mean intact

    with_std: bool, default=True
        If True, scale the data to unit variance. If False, keep the variance intact

    column_wise: bool, default=False
        If True, normalize each column separately. If False, normalize the whole
        matrix with respect to its total variance.

    rtol: float, default=0
        The relative tolerance for the optimization: variance is
        considered zero when it is less than abs(mean) * rtol + atol.

    atol: float, default=1.0E-12
        The relative tolerance for the optimization: variance is
        considered zero when it is less than abs(mean) * rtol + atol.

    copy : bool, default=None
        Copy the input X or not.

    Attributes
    ----------
    n_samples_seen_: int
        Number of samples in the reference ndarray

    n_features_: int
        Number of features in the reference ndarray

    mean_ : ndarray of shape (n_features,)
        The mean value for each feature in the training set.
        Equal to ndarray of zeros shape (n_features,) when ``with_mean=False``.

    scale_ : ndarray of shape (n_features,), float  or None
        The scaling factor, ndarray of shape (n_features,)
        when `column_wise=True` or float  when `column_wise = False`.

    copy : bool, default=None
        Copy the input X or not.

    Examples
    --------
    >>> import numpy as np
    >>> from skmatter.preprocessing import StandardFlexibleScaler
    >>> X = np.array([[ 1., -2.,  2.],
    ...               [-2.,  1.,  3.],
    ...               [ 4.,  1., -2.]])
    >>> transformer = StandardFlexibleScaler().fit(X)
    >>> transformer
    StandardFlexibleScaler()
    >>> transformer.transform(X)
    array([[ 0.        , -0.56195149,  0.28097574],
           [-0.84292723,  0.28097574,  0.56195149],
           [ 0.84292723,  0.28097574, -0.84292723]])
    >>> transformer.scale_ * transformer.transform(X)
    array([[ 0., -2.,  1.],
           [-3.,  1.,  2.],
           [ 3.,  1., -3.]])
    >>> transformer.scale_ * transformer.transform(X) + transformer.mean_
        array([[ 1., -2.,  2.],
               [-2.,  1.,  3.],
               [ 4.,  1., -2.]])
    """

    def __init__(
        self,
        with_mean=True,
        with_std=True,
        column_wise=False,
        rtol=0,
        atol=1e-12,
        copy=False,
    ):
        """Initialize StandardFlexibleScaler."""
        self.with_mean = with_mean
        self.with_std = with_std
        self.column_wise = column_wise
        self.rtol = rtol
        self.atol = atol
        self.n_samples_seen_ = 0
        self.copy = copy

    def fit(self, X, y=None, sample_weight=None):
        """Compute mean and scaling to be applied for subsequent normalization.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            The data used to compute the mean and standard deviation
            used for later scaling along the features axis.

        y: None
            Ignored.

        sample_weight: ndarray of shape (n_samples,)
            Weights for each sample. Sample weighting can be used to center
            (and scale) data using a weighted mean. Weights are internally
            normalized before preprocessing.

        Returns
        -------
        self : object
            Fitted scaler.
        """

        self.n_samples_seen_, self.n_features_ = X.shape

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=X.dtype)
            sample_weight = sample_weight / np.sum(sample_weight)

        if self.with_mean:
            self.mean_ = np.average(X, weights=sample_weight, axis=0)
        else:
            self.mean_ = np.zeros(self.n_features_)

        self.scale_ = 1.0
        if self.with_std:
            X_mean = np.average(X, weights=sample_weight, axis=0)
            var = np.average((X - X_mean) ** 2, weights=sample_weight, axis=0)

            if self.column_wise:
                if np.any(var < self.atol + abs(X_mean) * self.rtol):
                    raise ValueError("Cannot normalize a feature with zero variance")
                self.scale_ = np.sqrt(var)
            else:
                var_sum = var.sum()
                if var_sum < abs(np.mean(X_mean)) * self.rtol + self.atol:
                    raise ValueError("Cannot normalize a matrix with zero variance")
                self.scale_ = np.sqrt(var_sum)

        return self

    def transform(self, X, y=None, copy=None):
        """Normalize a vector based on previously computed mean and scaling.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data used to scale along the features axis.

        y: None
            Ignored.

        copy : bool, default=None
            Copy the input X or not.

        Returns
        -------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Transformed array.
        """

        copy = copy if copy is not None else self.copy
        X = self._validate_data(
            X,
            reset=False,
            accept_sparse="csr",
            copy=copy,
            estimator=self,
            dtype=FLOAT_DTYPES,
            force_all_finite="allow-nan",
        )
        check_is_fitted(self, attributes=["n_samples_seen_", "n_features_"])

        if self.n_features_ != X.shape[1]:
            raise ValueError("X shape does not match training shape")
        return (X - self.mean_) / self.scale_

    def inverse_transform(self, X_tr):
        """Scale back the data to the original representation

        Parameters
        ----------
        X_tr : ndarray of shape (n_samples, n_features)
            Transformed matrix

        Returns
        -------
        X : original matrix
        """

        check_is_fitted(self, attributes=["n_samples_seen_", "n_features_"])

        if self.n_features_ != X_tr.shape[1]:
            raise ValueError("X shape does not match training shape")
        return X_tr * self.scale_ + self.mean_


class KernelNormalizer(KernelCenterer):
    """Kernel centering method, similar to KernelCenterer,
    but with additional scaling and ability to pass a set of sample weights.

    Let K(x, z) be a kernel defined by phi(x)^T phi(z), where phi is a
    function mapping x to a Hilbert space. KernelNormalizer centers (i.e.,
    normalize to have zero mean) the data without explicitly computing phi(x).
    It is equivalent to centering and scaling phi(x) with
    sklearn.preprocessing.StandardScaler(with_std=False).

    Parameters
    ---------
    with_center: bool, default=True
        If True, center the kernel matrix before scaling. If False, do not
        center the kernel

    with_trace: bool, default=True
        If True, scale the kernel so that the trace is equal to the number of
        samples. If False, do not scale the kernel

    Attributes
    ----------
    K_fit_rows_ : ndarray of shape (n_samples,)
        Average of each column of kernel matrix.

    K_fit_all_ : float
        Average of kernel matrix.

    sample_weight_ : float
        Sample weights (if provided during the fit)

    scale_ : float
        Scaling parameter used when 'with_trace'=True
        Calculated as np.trace(K) / K.shape[0]

    Examples
    --------
    >>> from skmatter.preprocessing import KernelNormalizer
    >>> from sklearn.metrics.pairwise import pairwise_kernels
    >>> X = [[ 1., -2.,  2.],
    ...      [ -2.,  1.,  3.],
    ...      [ 4.,  1., -2.]]
    >>> K = pairwise_kernels(X, metric='linear')
    >>> K
    array([[  9.,   2.,  -2.],
           [  2.,  14., -13.],
           [ -2., -13.,  21.]])
    >>> transformer = KernelNormalizer().fit(K)
    >>> transformer
    KernelNormalizer()
    >>> transformer.transform(K)
    array([[ 0.39473684,  0.        , -0.39473684],
           [ 0.        ,  1.10526316, -1.10526316],
           [-0.39473684, -1.10526316,  1.5       ]])
    >>> transformer.scale_ * transformer.transform(K)
    array([[  5.,   0.,  -5.],
           [  0.,  14., -14.],
           [ -5., -14.,  19.]])
    >>>
    """

    def __init__(self, with_center=True, with_trace=True):
        """Initialize KernelNormalizer."""
        self.with_center = with_center
        self.with_trace = with_trace
        super().__init__()

    def fit(self, K=None, y=None, sample_weight=None):
        """Fit KernelFlexibleCenterer

        Parameters
        ----------
        K : ndarray of shape (n_samples, n_samples)
            Kernel matrix.

        y : None
            Ignored.

        sample_weight: ndarray of shape (n_samples,), default=None
            Weights for each sample. Sample weighting can be used to center (and
            scale) data using a weighted mean. Weights are internally normalized
            before preprocessing.

        Returns
        -------
        self : object
            Fitted transformer.
        """

        Kc = self._validate_data(K, copy=True, dtype=FLOAT_DTYPES, reset=False)

        if sample_weight is not None:
            self.sample_weight_ = _check_sample_weight(sample_weight, K, dtype=K.dtype)
            self.sample_weight_ = self.sample_weight_ / np.sum(self.sample_weight_)
        else:
            self.sample_weight_ = sample_weight

        if self.with_center:
            if self.sample_weight_ is not None:
                self.K_fit_rows_ = np.average(K, weights=self.sample_weight_, axis=0)
                self.K_fit_all_ = np.average(
                    self.K_fit_rows_, weights=self.sample_weight_
                )
            else:
                super().fit(K, y)

            K_pred_cols = np.average(Kc, weights=self.sample_weight_, axis=1)[
                :, np.newaxis
            ]
        else:
            self.K_fit_rows_ = np.zeros(Kc.shape[1])
            self.K_fit_all_ = 0.0
            K_pred_cols = np.zeros((Kc.shape[0], 1))

        if self.with_trace:
            Kc -= self.K_fit_rows_
            Kc -= K_pred_cols
            Kc += self.K_fit_all_

            self.scale_ = np.trace(Kc) / Kc.shape[0]
        else:
            self.scale_ = 1.0

        return self

    def transform(self, K, copy=True):
        """Center kernel matrix.

        Parameters
        ----------
        K : ndarray of shape (n_samples1, n_samples2)
            Kernel matrix.

        copy : bool, default=True
            Set to False to perform inplace computation.

        Returns
        -------
        K_new : ndarray of shape (n_samples1, n_samples2)
            Transformed array
        """

        check_is_fitted(self)
        K = self._validate_data(K, copy=copy, dtype=FLOAT_DTYPES, reset=False)

        if self.with_center:
            K_pred_cols = np.average(K, weights=self.sample_weight_, axis=1)[
                :, np.newaxis
            ]
        else:
            K_pred_cols = np.zeros((K.shape[0], 1))

        K -= self.K_fit_rows_
        K -= K_pred_cols
        K += self.K_fit_all_

        return K / self.scale_

    def fit_transform(self, K, y=None, sample_weight=None, copy=True, **fit_params):
        r"""Fit to data, then transform it.

        Parameters
        ----------
        K : ndarray of shape (n_samples, n_samples)
            Kernel matrix.

        y : None
            Ignored.

        sample_weight: ndarray of shape (n_samples,), default=None
            Weights for each sample. Sample weighting can be used to center (and
            scale) data using a weighted mean. Weights are internally normalized
            before preprocessing.

        \**fit_params:
            necessary for compatibility with the functions of the
            TransformerMixin class

        Returns
        -------
        K_new : ndarray of shape (n_samples1, n_samples2)
            Transformed array
        """
        self.fit(K, y, sample_weight=sample_weight)
        return self.transform(K, copy)


class SparseKernelCenterer(TransformerMixin, BaseEstimator):
    r"""Kernel centering method for sparse kernels, similar to
    KernelFlexibleCenterer.

    The main disadvantage of kernel methods, which is widely used in machine
    learning it is that they quickly grow in time and space complexity with the
    number of sample. It is clear that with a large dataset, not only do you
    need to store a huge amount of information, but you also need to use it
    constantly in calculations. In order to avoid this, so-called sparse kernel
    methods are used formulated from the low-dimensional (The Nystrom)
    approximation:

    .. math::
        \mathbf{K} \approx \hat{\mathbf{K}}_{N N}=\mathbf{K}_{N M} \mathbf{K}_{M M}^{-1} \mathbf{K}_{N M}^{T}

    where the subscripts for $\mathbf{K}$ denote the size of the sets of samples
    compared in each kernel, with $N$ being the size of the full data set and
    $M$ referring a small, active set containing $M$ samples. With this
    method it is only need to save and use the matrix $\mathbf{K}_{NM}$, i.e. it
    is possible to get a $N/M$ times improvement in the asymptotic by memory.

    Parameters
    ---------
    with_center: bool, default=True
        If True, center the kernel matrix before scaling. If False, do not
        center the kernel

    with_trace: bool, default=True
        If True, scale the kernel so that the trace is equal to the number of
        samples. If False, do not scale the kernel

    rcond: float, default 1E-12
        conditioning parameter to use when computing the Nystrom-approximated
        kernel for scaling

    Attributes
    ----------
    K_fit_rows_ : ndarray of shape (n_samples,)
        Average of each column of kernel matrix.

    K_fit_all_ : float
        Average of kernel matrix.

    sample_weight_ : float
        Sample weights (if provided during the fit)

    scale_ : float
        Scaling parameter used when 'with_trace'=True
        Calculated as np.trace(K) / K.shape[0]

    n_active_: int
        size of active set
    """

    def __init__(self, with_center=True, with_trace=True, rcond=1e-12):
        """Initialize SparseKernelCenterer."""

        self.with_center = with_center
        self.with_trace = with_trace
        self.rcond = rcond

    def fit(self, Knm, Kmm, y=None, sample_weight=None):
        """Fit KernelFlexibleCenterer

        Parameters
        ---------
        Knm: ndarray of shape (n_samples, n_active)
            Kernel matrix between the reference data set and the active set

        Kmm: ndarray of shape (n_active, n_active)
            Kernel matrix between the active set and itself

        y : None
            Ignored.

        sample_weight: ndarray of shape (n_samples,), default=None
            Weights for each sample. Sample weighting can be used to center (and
            scale) data using a weighted mean. Weights are internally normalized
            before preprocessing.

        Returns
        -------
        self : object
            Fitted transformer.
        """

        if Knm.shape[1] != Kmm.shape[0]:
            raise ValueError(
                "The reference kernel is not commensurate shape with the "
                "active kernel."
            )

        if Kmm.shape[0] != Kmm.shape[1]:
            raise ValueError("The active kernel is not square.")

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, Knm, dtype=Knm.dtype)
            sample_weight = sample_weight / np.sum(sample_weight)

        self.n_active_ = Kmm.shape[0]

        if self.with_center:
            self.K_fit_rows_ = np.average(Knm, weights=sample_weight, axis=0)
        else:
            self.K_fit_rows_ = np.zeros(Knm.shape[1])

        if self.with_trace:
            Knm_centered = Knm - self.K_fit_rows_

            Khat = Knm_centered @ np.linalg.pinv(Kmm, self.rcond) @ Knm_centered.T

            self.scale_ = np.sqrt(np.trace(Khat) / Knm.shape[0])
        else:
            self.scale_ = 1.0

        return self

    def transform(self, Knm, y=None):
        """Centering our Kernel. Previously you should fit data.

        Parameters
        ---------
        Knm: ndarray of shape (n_samples, n_active)
            Kernel matrix between the reference data set and the active set

        y : None
            Ignored.

        Returns
        -------
        K_new : ndarray of shape (n_samples, n_active)
            Transformed array
        """
        check_is_fitted(self, attributes=["scale_", "K_fit_rows_", "n_active_"])

        if Knm.shape[1] != self.n_active_:
            raise ValueError(
                "The reference kernel and received kernel have different shape"
            )

        Kc = (Knm - self.K_fit_rows_) / self.scale_

        return Kc

    def fit_transform(self, Knm, Kmm, y=None, sample_weight=None, **fit_params):
        r"""Fit to data, then transform it.

        Parameters
        ---------
        Knm: ndarray of shape (n_samples, n_active)
            Kernel matrix between the reference data set and the active set

        Kmm: ndarray of shape (n_active, n_active)
            Kernel matrix between the active set and itself

        y : None
            Ignored.

        sample_weight: ndarray of shape (n_samples,), default=None
            Weights for each sample. Sample weighting can be used to center (and
            scale) data using a weighted mean. Weights are internally normalized
            before preprocessing.

        \**fit_params:
            necessary for compatibility with the functions of the
            TransformerMixin class

        Returns
        -------
        K_new : ndarray of shape (n_samples, n_active)
            Transformed array
        """
        self.fit(Knm=Knm, Kmm=Kmm, sample_weight=sample_weight)
        return self.transform(Knm)



"""
The :mod:`sklearn.sample selection` module will include FPS and CUR selection, each
with the optional PCov-flavor
"""

from ._base import (
    CUR,
    FPS,
    DirectionalConvexHull,
    PCovCUR,
    PCovFPS,
)
from ._voronoi_fps import VoronoiFPS

__all__ = ["PCovFPS", "PCovCUR", "FPS", "CUR", "DirectionalConvexHull", "VoronoiFPS"]



import unittest

import numpy as np
from sklearn.datasets import load_diabetes as get_dataset
from sklearn.exceptions import NotFittedError

from skmatter._selection import GreedySelector


class GreedyTester(GreedySelector):
    def __init__(
        self, n_to_select=None, score_threshold=None, selection_type="feature", **kwargs
    ):
        super().__init__(
            selection_type=selection_type,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            **kwargs,
        )

    def score(self, X, y=None):
        scores = np.linalg.norm(X, axis=0)
        scores[self.selected_idx_] = 0.0
        return scores


class TestGreedy(unittest.TestCase):
    def setUp(self):
        self.X, _ = get_dataset(return_X_y=True)

    def test_bad_type(self):
        with self.assertRaises(
            ValueError, msg="Only feature and sample selection supported."
        ):
            _ = GreedyTester(selection_type="bad").fit(self.X)

    def test_score_threshold(self):
        selector = GreedyTester(score_threshold=200, n_to_select=7)
        with self.assertWarns(
            Warning, msg="Score threshold of 200 reached. Terminating search at 6 / 7."
        ):
            selector.fit(self.X)

    def test_score_threshold_and_full(self):
        with self.assertRaises(ValueError) as cm:
            _ = GreedyTester(score_threshold=20, full=True, n_to_select=12).fit(self.X)
        self.assertEqual(
            str(cm.exception),
            "You cannot specify both `score_threshold` and `full=True`.",
        )

    def test_bad_warm_start(self):
        selector = GreedyTester()
        with self.assertRaises(ValueError) as cm:
            selector.fit(self.X, warm_start=True)
        self.assertTrue(
            str(cm.exception),
            "Cannot fit with warm_start=True without having been previously initialized",
        )

    def test_bad_y(self):
        self.X, self.Y = get_dataset(return_X_y=True)
        selector = GreedyTester(n_to_select=2)
        with self.assertRaises(ValueError):
            selector.fit(X=self.X, y=self.Y[:2])

    def test_bad_transform(self):
        selector = GreedyTester(n_to_select=2)
        selector.fit(self.X)
        with self.assertRaises(ValueError) as cm:
            _ = selector.transform(self.X[:, :3])
        self.assertEqual(
            str(cm.exception), "X has a different shape than during fitting."
        )

    def test_no_nfeatures(self):
        selector = GreedyTester()
        selector.fit(self.X)
        self.assertEqual(len(selector.selected_idx_), self.X.shape[1] // 2)

    def test_decimal_nfeatures(self):
        selector = GreedyTester(n_to_select=0.2)
        selector.fit(self.X)
        self.assertEqual(len(selector.selected_idx_), int(self.X.shape[1] * 0.2))

    def test_bad_nfeatures(self):
        for nf in [1.2, "1", 20]:
            with self.subTest(n_features=nf):
                selector = GreedyTester(n_to_select=nf)
                with self.assertRaises(ValueError) as cm:
                    selector.fit(self.X)
                self.assertEqual(
                    str(cm.exception),
                    (
                        "n_to_select must be either None, an integer in "
                        "[1, n_features] representing the absolute number "
                        "of features, or a float in (0, 1] representing a "
                        f"percentage of features to select. Got {nf} "
                        f"features and an input with {self.X.shape[1]} feature."
                    ),
                )

    def test_not_fitted(self):
        with self.assertRaises(NotFittedError):
            selector = GreedyTester()
            _ = selector._get_support_mask()

    def test_fitted(self):
        selector = GreedyTester()
        selector.fit(self.X)
        _ = selector._get_support_mask()

        Xr = selector.transform(self.X)
        self.assertEqual(Xr.shape[1], self.X.shape[1] // 2)

    def test_size_input(self):
        X = np.array([1, 2, 3, 4, 5]).reshape(-1, 1)
        selector_sample = GreedyTester(selection_type="sample")
        selector_feature = GreedyTester(selection_type="feature")
        selector_sample.fit(X)
        with self.assertRaises(ValueError) as cm:
            selector_feature.fit(X)
        self.assertEqual(
            str(cm.exception),
            "Found array with 1 feature(s) (shape=(5, 1)) while a minimum of 2 is required.",
        )
        X = X.reshape(1, -1)
        selector_feature.fit(X)
        with self.assertRaises(ValueError) as cm:
            selector_sample.fit(X)
        self.assertEqual(
            str(cm.exception),
            "Found array with 1 sample(s) (shape=(1, 5)) while a minimum of 2 is required.",
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest

from sklearn.datasets import load_iris
import skmatter.model_selection
import sklearn.model_selection


class SplitTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.X = load_iris().data[:10]
        cls.seed = 0x5F3759DF

    def test_train_test_splits(self):
        # see if train_test_split of skmatter agrees with the one of sklearn
        sklearn_outputs = sklearn.model_selection.train_test_split(
            self.X, random_state=self.seed
        )
        skmatter_outputs = skmatter.model_selection.train_test_split(
            self.X, random_state=self.seed
        )
        for i in range(len(skmatter_outputs)):
            self.assertTrue((sklearn_outputs[i] == skmatter_outputs[i]).all())

    def test_train_test_splits_train_test_overlap(self):
        # tests that a test/train split which necessitates overlap returns the right number of points in each set
        X_train, X_test = skmatter.model_selection.train_test_split(
            self.X,
            train_size=0.8,
            test_size=0.8,
            train_test_overlap=True,
            random_state=self.seed,
        )
        self.assertTrue(len(X_train) == len(X_test) == int(0.8 * self.X.shape[0]))

    def test_train_test_splits_train_test_overlap_full_test_set(self):
        # tests that the entire dataset can be used as the testing set
        X_train, X_test = skmatter.model_selection.train_test_split(
            self.X,
            train_size=0.8,
            test_size=1.0,
            train_test_overlap=True,
            random_state=self.seed,
        )
        self.assertTrue((self.X == X_test).all())

    def test_train_test_splits_train_test_overlap_full_train_test_set(self):
        # tests that the full dataset can be "split" to both train and test set
        X_train, X_test = skmatter.model_selection.train_test_split(
            self.X,
            train_size=1.0,
            test_size=1.0,
            train_test_overlap=True,
            random_state=self.seed,
        )
        self.assertTrue((X_train == X_test).all())


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from sklearn.datasets import load_iris
from sklearn.utils import (
    check_random_state,
    extmath,
)

from skmatter.metrics import (
    global_reconstruction_distortion,
    global_reconstruction_error,
    local_reconstruction_error,
    pointwise_local_reconstruction_error,
)


class ReconstructionMeasuresTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        features = load_iris().data
        cls.features_small = features[:20, [0, 1]]
        cls.features_large = features[:20, [0, 1, 0, 1]]
        cls.eps = 1e-5
        cls.n_local_points = 15

        random_state = 0
        random_state = check_random_state(random_state)
        random_orthonormal_mat = extmath.randomized_range_finder(
            np.eye(cls.features_small.shape[1]),
            size=cls.features_small.shape[1],
            n_iter=10,
            random_state=random_state,
        )
        cls.features_rotated_small = cls.features_small @ random_orthonormal_mat

    def test_global_reconstruction_error_identity(self):
        gfre_val = global_reconstruction_error(self.features_large, self.features_large)
        self.assertTrue(
            abs(gfre_val) < self.eps,
            f"global_reconstruction_error {gfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_global_reconstruction_error_small_to_large(self):
        # tests that the GRE of a small set of features onto a larger set of features returns within a threshold of zero
        gfre_val = global_reconstruction_error(self.features_small, self.features_large)
        self.assertTrue(
            abs(gfre_val) < self.eps,
            f"global_reconstruction_error {gfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_global_reconstruction_error_large_to_small(self):
        # tests that the GRE of a large set of features onto a smaller set of features returns within a threshold of zero
        gfre_val = global_reconstruction_error(self.features_large, self.features_small)
        self.assertTrue(
            abs(gfre_val) < self.eps,
            f"global_reconstruction_error {gfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_global_reconstruction_distortion_identity(self):
        # tests that the GRD of a set of features onto itself returns within a threshold of zero
        gfrd_val = global_reconstruction_distortion(
            self.features_large, self.features_large
        )
        self.assertTrue(
            abs(gfrd_val) < self.eps,
            f"global_reconstruction_error {gfrd_val} surpasses threshold for zero {self.eps}",
        )

    def test_global_reconstruction_distortion_small_to_large(self):
        # tests that the GRD of a small set of features onto a larger set of features returns within a threshold of zero
        # should just run
        global_reconstruction_error(self.features_small, self.features_large)

    def test_global_reconstruction_distortion_large_to_small(self):
        # tests that the GRD of a large set of features onto a smaller set of features returns within a threshold of zero
        # should just run
        global_reconstruction_error(self.features_large, self.features_small)

    def test_global_reconstruction_distortion_small_to_rotated_small(self):
        # tests that the GRD of a small set of features onto a rotation of itself returns within a threshold of zero
        gfrd_val = global_reconstruction_distortion(
            self.features_small, self.features_rotated_small
        )
        self.assertTrue(
            abs(gfrd_val) < self.eps,
            f"global_reconstruction_error {gfrd_val} surpasses threshold for zero {self.eps}",
        )

    def test_local_reconstruction_error_identity(self):
        # tests that the local reconstruction error of a set of features onto itself returns within a threshold of zero

        lfre_val = local_reconstruction_error(
            self.features_large, self.features_large, self.n_local_points
        )
        self.assertTrue(
            abs(lfre_val) < self.eps,
            f"local_reconstruction_error {lfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_local_reconstruction_error_small_to_large(self):
        # tests that the local reconstruction error of a small set of features onto a larger set of features returns within a threshold of zero

        lfre_val = local_reconstruction_error(
            self.features_small, self.features_large, self.n_local_points
        )
        self.assertTrue(
            abs(lfre_val) < self.eps,
            f"local_reconstruction_error {lfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_local_reconstruction_error_large_to_small(self):
        # tests that the local reconstruction error of a larger set of features onto a smaller set of features returns within a threshold of zero

        lfre_val = local_reconstruction_error(
            self.features_large, self.features_small, self.n_local_points
        )
        self.assertTrue(
            abs(lfre_val) < self.eps,
            f"local_reconstruction_error {lfre_val} surpasses threshold for zero {self.eps}",
        )

    def test_local_reconstruction_error_train_idx(self):
        # tests that the local reconstruction error works when specifying a manual train idx

        lfre_val = pointwise_local_reconstruction_error(
            self.features_large,
            self.features_large,
            self.n_local_points,
            train_idx=np.arange((len(self.features_large) // 4)),
        )
        test_size = len(self.features_large) - (len(self.features_large) // 4)
        self.assertTrue(
            len(lfre_val) == test_size,
            f"size of pointwise LFRE  {len(lfre_val)} differs from expected test set size {test_size}",
        )

    def test_local_reconstruction_error_test_idx(self):
        # tests that the local reconstruction error works when specifying a manual train idx

        lfre_val = pointwise_local_reconstruction_error(
            self.features_large,
            self.features_large,
            self.n_local_points,
            test_idx=np.arange((len(self.features_large) // 4)),
        )
        test_size = len(self.features_large) // 4
        self.assertTrue(
            len(lfre_val) == test_size,
            f"size of pointwise LFRE  {len(lfre_val)} differs from expected test set size {test_size}",
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from setuptools import setup
import re

__version__ = re.search(
    r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', open("skmatter/__init__.py").read()
).group(1)

if __name__ == "__main__":
    setup(version=__version__)



from copy import deepcopy

import numpy as np
from sklearn.base import clone
from sklearn.exceptions import NotFittedError
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.utils.extmath import randomized_svd
from sklearn.utils.validation import check_is_fitted


def check_lr_fit(regressor, X, y):
    r"""
    Checks that a (linear) regressor is fitted, and if not,
    fits it with the provided data

    :param regressor: sklearn-style regressor
    :type regressor: object
    :param X: feature matrix with which to fit the regressor
        if it is not already fitted
    :type X: array
    :param y: target values with which to fit the regressor
        if it is not already fitted
    :type y: array
    """
    try:
        check_is_fitted(regressor)
        fitted_regressor = deepcopy(regressor)

        # Check compatibility with X
        fitted_regressor._validate_data(X, y, reset=False, multi_output=True)

        # Check compatibility with y
        if fitted_regressor.coef_.ndim != y.ndim:
            raise ValueError(
                "The regressor coefficients have a dimension incompatible "
                "with the supplied target space. "
                "The coefficients have dimension %d and the targets "
                "have dimension %d" % (fitted_regressor.coef_.ndim, y.ndim)
            )
        elif y.ndim == 2:
            if fitted_regressor.coef_.shape[0] != y.shape[1]:
                raise ValueError(
                    "The regressor coefficients have a shape incompatible "
                    "with the supplied target space. "
                    "The coefficients have shape %r and the targets "
                    "have shape %r" % (fitted_regressor.coef_.shape, y.shape)
                )

    except NotFittedError:
        fitted_regressor = clone(regressor)
        fitted_regressor.fit(X, y=y)

    return fitted_regressor


def check_krr_fit(regressor, K, X, y):
    r"""
    Checks that a (kernel ridge) regressor is fitted, and if not,
    fits it with the provided data

    :param regressor: sklearn-style regressor
    :type regressor: object
    :param K: kernel matrix with which to fit the regressor
        if it is not already fitted
    :type K: array
    :param X: feature matrix with which to check the regressor
    :type X: array
    :param y: target values with which to fit the regressor
        if it is not already fitted
    :type y: array
    """
    try:
        check_is_fitted(regressor)
        fitted_regressor = deepcopy(regressor)

        # Check compatibility with K
        fitted_regressor._validate_data(X, y, reset=False, multi_output=True)

        # Check compatibility with y
        if fitted_regressor.dual_coef_.ndim != y.ndim:
            raise ValueError(
                "The regressor coefficients have a dimension incompatible "
                "with the supplied target space. "
                "The coefficients have dimension %d and the targets "
                "have dimension %d" % (fitted_regressor.dual_coef_.ndim, y.ndim)
            )
        elif y.ndim == 2:
            if fitted_regressor.dual_coef_.shape[1] != y.shape[1]:
                raise ValueError(
                    "The regressor coefficients have a shape incompatible "
                    "with the supplied target space. "
                    "The coefficients have shape %r and the targets "
                    "have shape %r" % (fitted_regressor.dual_coef_.shape, y.shape)
                )

    except NotFittedError:
        fitted_regressor = clone(regressor)

        # Use a precomputed kernel
        # to avoid re-computing K
        fitted_regressor.set_params(kernel="precomputed")
        fitted_regressor.fit(K, y=y)

    return fitted_regressor


def pcovr_covariance(
    mixing,
    X,
    Y,
    rcond=1e-12,
    return_isqrt=False,
    rank=None,
    random_state=0,
    iterated_power="auto",
):
    r"""
    Creates the PCovR modified covariance

    .. math::

        \mathbf{\tilde{C}} = \alpha \mathbf{X}^T \mathbf{X} +
        (1 - \alpha) \left(\left(\mathbf{X}^T
        \mathbf{X}\right)^{-\frac{1}{2}} \mathbf{X}^T
        \mathbf{\hat{Y}}\mathbf{\hat{Y}}^T \mathbf{X} \left(\mathbf{X}^T
        \mathbf{X}\right)^{-\frac{1}{2}}\right)

    where :math:`\mathbf{\hat{Y}}`` are the properties obtained by linear regression.

    Parameters
    ----------
    mixing : float
             mixing parameter, as described in PCovR as :math:`{\alpha}`,

    X : ndarray of shape (n x m)
        Data matrix :math:`\mathbf{X}`

    Y : ndarray of shape (n x p)
        Array to include in biased selection when mixing < 1

    rcond : float,  default=1E-12
            threshold below which eigenvalues will be considered 0,

    return_isqrt : bool, default=False
                    Whether to return the calculated inverse square root of
                    the covariance. Used when inverse square root is needed
                    and the pcovr_covariance has already been calculated

    rank : int, default=min(X.shape)
           number of eigenpairs to estimate the inverse square root with

    random_state : int, default=0
                   random seed to use for randomized svd

    """

    C = np.zeros((X.shape[1], X.shape[1]), dtype=np.float64)

    if mixing < 1 or return_isqrt:
        if rank is None:
            rank = min(X.shape)

        if rank >= min(X.shape):
            vC, UC = np.linalg.eigh(X.T @ X)

            vC = np.flip(vC)
            UC = np.flip(UC, axis=1)[:, vC > rcond]
            vC = np.sqrt(vC[vC > rcond])

        else:
            _, vC, UC = randomized_svd(
                X,
                n_components=rank,
                n_iter=iterated_power,
                flip_sign=True,
                random_state=random_state,
            )

            UC = UC.T[:, (vC**2) > rcond]
            vC = vC[(vC**2) > rcond]

        C_isqrt = UC @ np.diagflat(1.0 / vC) @ UC.T

        # parentheses speed up calculation greatly
        C_Y = C_isqrt @ (X.T @ Y)
        C_Y = C_Y.reshape((C.shape[0], -1))
        C_Y = np.real(C_Y)

        C += (1 - mixing) * C_Y @ C_Y.T

    if mixing > 0:
        C += (mixing) * (X.T @ X)

    if return_isqrt:
        return C, C_isqrt
    else:
        return C


def pcovr_kernel(mixing, X, Y, **kernel_params):
    r"""
    Creates the PCovR modified kernel distances

    .. math::

        \mathbf{\tilde{K}} = \alpha \mathbf{K} +
        (1 - \alpha) \mathbf{Y}\mathbf{Y}^T

    the default kernel is the linear kernel, such that:

    .. math::

        \mathbf{\tilde{K}} = \alpha \mathbf{X} \mathbf{X}^T +
        (1 - \alpha) \mathbf{Y}\mathbf{Y}^T

    Parameters
    ----------

    mixing : float
             mixing parameter, as described in PCovR as :math:`{\alpha}`

    X : ndarray of shape (n x m)
        Data matrix :math:`\mathbf{X}`

    Y : ndarray of shape (n x p)
        Array to include in biased selection when mixing < 1

    kernel_params : dict, optional
                    dictionary of arguments to pass to pairwise_kernels
                    if none are specified, assumes that the kernel is linear

    """

    K = np.zeros((X.shape[0], X.shape[0]))
    if mixing < 1:
        K += (1 - mixing) * Y @ Y.T
    if mixing > 0:
        if "kernel" not in kernel_params:
            K += (mixing) * X @ X.T
        elif kernel_params.get("kernel") != "precomputed":
            K += (mixing) * pairwise_kernels(X, **kernel_params)
        else:
            K += (mixing) * X

    return K



"""
Sequential feature selection
"""

from .._selection import (
    _CUR,
    _FPS,
    _PCovCUR,
    _PCovFPS,
)


class FPS(_FPS):
    """
    Transformer that performs Greedy Feature Selection using Farthest Point Sampling.

    Parameters
    ----------

    initialize: int, list of int, or 'random', default=0
        Index of the first selection(s). If 'random', picks a random
        value when fit starts. Stored in :py:attr:`self.initialize`.

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining features. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected features, for use in fitting

    """

    def __init__(
        self,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="feature",
            initialize=initialize,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class PCovFPS(_PCovFPS):
    """Transformer that performs Greedy Feature Selection using PCovR-weighted
    Farthest Point Sampling.

    Parameters
    ----------

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`

    initialize: int or 'random', default=0
        Index of the first selection. If 'random', picks a random
        value when fit starts.

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining features. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected features, for use in fitting

    """

    def __init__(
        self,
        mixing=0.5,
        initialize=0,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="feature",
            mixing=mixing,
            initialize=initialize,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class CUR(_CUR):
    """Transformer that performs Greedy Feature Selection by choosing features
    which maximize the magnitude of the right singular vectors, consistent with
    classic CUR matrix decomposition.

    Parameters
    ----------
    recompute_every : int
                      number of steps after which to recompute the pi score
                      defaults to 1, if 0 no re-computation is done

    k : int
        number of eigenvectors to compute the importance score with, defaults to 1

    tolerance: float
         threshold below which scores will be considered 0, defaults to 1E-12


    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining features. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    X_current_ : ndarray (n_samples, n_features)
                  The original matrix orthogonalized by previous selections

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected features, for use in fitting

    """

    def __init__(
        self,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="feature",
            recompute_every=recompute_every,
            k=k,
            tolerance=tolerance,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )


class PCovCUR(_PCovCUR):
    """Transformer that performs Greedy Feature Selection by choosing features
    which maximize the importance score :math:`\\pi`, which is the sum over
    the squares of the first :math:`k` components of the PCovR-modified
    right singular vectors.

    Parameters
    ----------
    recompute_every : int
                      number of steps after which to recompute the pi score
                      defaults to 1, if 0 no re-computation is done

    k : int
        number of eigenvectors to compute the importance score with, defaults to 1

    tolerance: float
         threshold below which scores will be considered 0, defaults to 1E-12

    mixing: float, default=0.5
            The PCovR mixing parameter, as described in PCovR as
            :math:`{\\alpha}`. Stored in :py:attr:`self.mixing`.

    n_to_select : int or float, default=None
        The number of selections to make. If `None`, half of the features are
        selected. If integer, the parameter is the absolute number of selections
        to make. If float between 0 and 1, it is the fraction of the total dataset to
        select. Stored in :py:attr:`self.n_to_select`.

    score_threshold : float, default=None
        Threshold for the score. If `None` selection will continue until the
        n_to_select is chosen. Otherwise will stop when the score falls below the threshold.
        Stored in :py:attr:`self.score_threshold`.

    score_threshold_type : str, default="absolute"
        How to interpret the ``score_threshold``. When "absolute", the score used by
        the selector is compared to the threshold directly. When "relative", at each iteration,
        the score used by the selector is compared proportionally to the score of the first
        selection, i.e. the selector quits when ``current_score / first_score < threshold``.
        Stored in :py:attr:`self.score_threshold_type`.

    progress_bar: bool, default=False
              option to use `tqdm <https://tqdm.github.io/>`_
              progress bar to monitor selections. Stored in :py:attr:`self.report_progress`.

    full : bool, default=False
        In the case that all non-redundant selections are exhausted, choose
        randomly from the remaining features. Stored in :py:attr:`self.full`.

    random_state: int or RandomState instance, default=0

    Attributes
    ----------

    X_current_ : ndarray (n_samples, n_features)
                  The original matrix orthogonalized by previous selections

    y_current_ : ndarray (n_samples, n_properties)
                The targets orthogonalized by a regression on
                the previous selections.

    n_selected_ : int
                  Counter tracking the number of selections that have been made
    X_selected_ : ndarray,
                  Matrix containing the selected features, for use in fitting

    """

    def __init__(
        self,
        mixing=0.5,
        recompute_every=1,
        k=1,
        tolerance=1e-12,
        n_to_select=None,
        score_threshold=None,
        score_threshold_type="absolute",
        progress_bar=False,
        full=False,
        random_state=0,
    ):
        super().__init__(
            selection_type="feature",
            mixing=mixing,
            recompute_every=recompute_every,
            k=k,
            tolerance=tolerance,
            n_to_select=n_to_select,
            score_threshold=score_threshold,
            score_threshold_type=score_threshold_type,
            progress_bar=progress_bar,
            full=full,
            random_state=random_state,
        )
